"""Splits Brookhaven files up into seperate scan files and dumps them in a folder 'BNLSplitFiles'.

R.C.Temple

Will overwrite any files already split but that should be ok for an update.
"""

# pylint: disable=invalid-name
import os
import numpy as np

import Stoner

# Get file open ###############
while True:
    try:
        directory = input(
            "Enter the directory path where your data is stored:\n"
        )
        os.chdir(directory)
        filename = input(
            "Enter the filename (including extension) for your file\r\n"
        )
        mainFP = open(filename, "r", encoding="utf-8")
        break
    except IOError:
        print("Oops I couldn't find that file.")
if "BNLSplitFiles" not in os.listdir(directory):
    os.mkdir("BNLSplitFiles")
os.chdir("BNLSplitFiles")

# Main algorithm ###########

# writeName=re.split(r'[.]',filename)
writeFP = open(
    "title.txt", "w", encoding="utf-8"
)  # title sequence goes in this file
counter = 1  # this will label the files
for line in mainFP:
    if line[0:2] == "#S":
        if int(line.split()[1]) != counter:
            raise ValueError  # check for inconsistencies with filenames and scan numbers
        writeFP.close()
        writeFP = open(str(counter) + ".bnl", "w", encoding="utf-8")
        counter += 1
    if line[0:2] != "#C":
        writeFP.write(line)
        # ignore #C statements which are usually abort and rarely useful, they come
        # after data and before the next #S"""
writeFP.close()
mainFP.close()

# test files
filelist = os.listdir(os.getcwd())
filelist.pop(0)
print("Testing files with Stoner:")
for filename in filelist:
    if filename.split(".")[-1] == "bnl":
        d = Stoner.formats.facilities.BNLFile(
            filename
        )  # will throw suitable errors if there are problems
        if len(np.shape(d.data)) == 1:
            print("Removing file {} due to lack of data".format(filename))
            d = 0
            os.remove(
                filename
            )  # delete files with only 1 dimensional data (or with
            # no data), they'll cause problems later
            continue
        print("{} OK".format(filename))
print("Done.")



# -*- coding: utf-8 -*-
"""
Test the Image Widgets used for selections

@author: phygbu
"""
import os
import threading
import time

import numpy as np
from matplotlib.backend_bases import Event

import pytest
import Stoner
from Stoner.Image.widgets import send_event as _event


def _trigger(image):
    time.sleep(5)
    _event(image, "on_click", xdata=1, ydata=1, button=1)
    for coord in np.linspace(1, 100, 51):
        _event(image, "draw_line", xdata=coord, ydata=coord)
    _event(image, "on_click", xdata=coord, ydata=coord, button=1)


def _trigger2(image):
    time.sleep(5)
    _event(image, "keypress", xdata=50, ydata=75, key="x")
    _event(image, "on_click", xdata=50, ydata=75, button=1)


def _trigger3(image):
    time.sleep(5)
    _event(image, "keypress", xdata=50, ydata=75, key="y")
    _event(image, "on_click", xdata=50, ydata=75, button=1)


def _trigger4(image):
    time.sleep(5)
    select = image._image._select
    event1 = Event("fake", select.fig.canvas)
    event1.xdata = 25
    event1.ydata = 25
    event2 = Event("fake", select.fig.canvas)
    event2.xdata = 75
    event2.ydata = 50
    select.on_select(event1, event2)
    time.sleep(2)
    _event(image, "finish", key="enter")


def _trigger5(image, mode):
    time.sleep(5)
    _event(image, ["draw", "on_click"], xdata=50, ydata=25, button=1, dblclick=False)
    _event(image, ["draw", "on_click"], xdata=75, ydata=50, button=1, dblclick=False)
    _event(image, "keypress", xdata=50, ydata=75, key=mode.lower()[0])
    if mode == "c":  # add some extra points:
        _event(image, ["draw", "on_click"], xdata=30, ydata=40, button=1, dblclick=False)
        _event(image, ["draw", "on_click"], xdata=30, ydata=30, button=1, dblclick=False)
    _event(image, "keypress", xdata=50, ydata=75, key="i")
    time.sleep(2)
    _event(image, "keypress", xdata=50, ydata=75, key="enter")


def _trigger6(image, mode):
    time.sleep(5)
    _event(image, ["draw", "on_click"], xdata=50, ydata=25, button=1, dblclick=False)
    _event(image, ["draw", "on_click"], xdata=75, ydata=50, button=1, dblclick=False)
    time.sleep(2)
    _event(image, "keypress", xdata=50, ydata=75, key="escape")


def test_profile_line():
    os.chdir(Stoner.__homepath__ / ".." / "sample-data")
    img = Stoner.HDF5.STXMImage("Sample_Image_2017-10-15_100.hdf5")
    thread = threading.Thread(target=_trigger, args=(img,))
    thread.start()
    result = img.profile_line()
    assert len(result) == 142
    assert result.x.min() == 0.0
    assert np.isclose(result.x.max(), 140, atol=0.01)
    assert np.isclose(result.y.mean(), 26548.72, atol=0.01)
    thread = threading.Thread(target=_trigger2, args=(img,))
    thread.start()
    result = img.profile_line()
    assert len(result) == 101
    assert result.x.min() == 0.0
    assert result.x.max() == 100.0
    assert np.isclose(result.y.mean(), 20022.16, atol=0.01)
    thread = threading.Thread(target=_trigger3, args=(img,))
    thread.start()
    result = img.profile_line()
    assert len(result) == 101
    assert result.x.min() == 0.0
    assert result.x.max() == 100.0
    assert np.isclose(result.y.mean(), 27029.16, atol=0.01)


def test_crop_with_ui():
    os.chdir(Stoner.__homepath__ / ".." / "sample-data")
    img = Stoner.HDF5.STXMImage("Sample_Image_2017-10-15_100.hdf5")
    thread = threading.Thread(target=_trigger4, args=(img,))
    thread.start()
    result = img.crop()
    assert result.shape == (25, 50)


def test_mask_select():
    os.chdir(Stoner.__homepath__ / ".." / "sample-data")
    img = Stoner.HDF5.STXMImage("Sample_Image_2017-10-15_100.hdf5")
    thread = threading.Thread(target=_trigger5, args=(img, "p"))
    thread.start()
    img.mask.select()
    result = img.mask.sum()
    assert result == 9324, f"Mask selection by polygon failed result={result}"
    img.mask = False
    thread = threading.Thread(target=_trigger5, args=(img, "c"))
    thread.start()
    img.mask.select()
    result = img.mask.sum()
    assert result in (3664, 7489), f"Mask selection by circle failed result={result}"
    img.mask = False
    thread = threading.Thread(target=_trigger5, args=(img, "r"))
    thread.start()
    img.mask.select()
    result = img.mask.sum()
    assert result == 8699, f"Mask selection by reverse failed result={result}"
    img.mask = False
    thread = threading.Thread(target=_trigger6, args=(img, "c"))
    thread.start()
    img.mask.select()
    result = img.mask.sum()
    assert result == 0, f"Cancelling selkection failed with result={result}"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Module to work with scan files from an AttocubeSPM running Daisy."""
__all__ = ["AttocubeScan"]
import pathlib
from os import path
from copy import deepcopy
from glob import glob
import re
import importlib

from numpy import genfromtxt, linspace, meshgrid, array, prod
from scipy.interpolate import griddata
from scipy.optimize import curve_fit
import h5py

from ..compat import string_types, bytes2str
from ..core.base import typeHintedDict
from ..Image import ImageStack, ImageFile, ImageArray
from ..HDF5 import HDFFileManager
from ..tools.file import FileManager
from ..core.exceptions import StonerLoadError

PARAM_RE = re.compile(r"^([\d\\.eE\+\-]+)\s*([\%A-Za-z]\S*)?$")
SCAN_NO = re.compile(r"SC_(\d+)")


def parabola(X, cx, cy, a, b, c):
    """Parabola in the X-Y plane for levelling an image."""
    x, y = X
    return a * (x - cx) ** 2 + b * (y - cy) ** 2 + c


def plane(X, a, b, c):
    """Plane equation for levelling an image."""
    x, y = X
    return a * x + b * y + c


def _raise_error(openfile, message=""):
    """Raise a StonerLoadError after trying to close file."""
    try:
        raise StonerLoadError(message)
    finally:
        try:
            openfile.close()
        except (AttributeError, TypeError, ValueError, IOError):
            pass


class AttocubeScanMixin:
    """Provides the specialist methods for dealing with Attocube SPM scan files.

    See :py:class:`AttocubeScan` for details."""

    def __init__(self, *args, **kargs):
        """Construct the attocube subclass of ImageStack."""
        args = list(args)
        if len(args) > 0:
            for ix, arg in enumerate(args):
                if isinstance(arg, pathlib.PurePath):
                    args[ix] = str(arg)
        if len(args) > 0 and isinstance(args[0], string_types):
            root_name = args.pop(0)
            scan = SCAN_NO.search(root_name)
            if scan:
                scan = int(scan.groups()[0])
            else:
                scan = -1
        elif len(args) > 0 and isinstance(args[0], int):
            scan = args.pop(0)
            root_name = f"SC_{scan:03d}"
        else:
            root_name = kargs.pop("root", None)
            scan = kargs.pop("scan", -1)

        regrid = kargs.pop("regrid", False)

        super().__init__(*args, **kargs)

        self._common_metadata = typeHintedDict()

        if root_name:
            self._load_files(root_name, regrid)

        self.scan_no = scan

        self._common_metadata["Scan #"] = scan

        self.compression = "gzip"
        self.compression_opts = 6

    def __clone__(self, other=None, attrs_only=False):
        """Do whatever is necessary to copy attributes from self to other.

        Note:
            We're in the base class here, so we don't call super() if we can't handle this, then we're stuffed!


        """
        other = super().__clone__(other, attrs_only)
        other._common_metadata = deepcopy(self._common_metadata)
        return other

    def __getitem__(self, name):
        """Allow scans to be got by channel name as well as normal indexing."""
        if isinstance(name, string_types):
            for ix, ch in enumerate(self.channels):
                if name in ch:
                    return self[ix]
        return super().__getitem__(name)

    @property
    def channels(self):
        """Get the list of channels saved in the scan."""
        if len(self) > 0:
            return self.metadata.slice("display", values_only=True)
        return []

    def _load(self, filename, *args, **kargs):
        """Load data from a hdf5 file.

        Args:
            h5file (string or h5py.Group):
                Either a string or an h5py Group object to load data from

        Returns:
            itself after having loaded the data
        """
        if filename is None or not filename:
            self.get_filename("r")
            filename = self.filename
        else:
            self.filename = filename
        with HDFFileManager(self.filename, "r") as f:
            if "type" not in f.attrs:
                _raise_error(f, message="HDF5 Group does not specify the type attribute used to check we can load it.")
            typ = bytes2str(f.attrs["type"])
            if typ != type(self).__name__ and "module" not in f.attrs:
                _raise_error(
                    f,
                    message=f"HDF5 Group is not a {type(self).__name__} and does not specify a module to use to load.",
                )
            loader = None
            if typ == type(self).__name__:
                loader = getattr(type(self), "read_hdf5")
            else:
                mod = importlib.import_module(bytes2str(f.attrs["module"]))
                cls = getattr(mod, typ)
                loader = getattr(cls, "read_hdf5")
            if loader is None:
                raise StonerLoadError("Could not et loader for {bytes2str(f.attrs['module'])}.{typ}")

        return loader(f, *args, **kargs)

    def _instantiate(self, idx):
        """Reconstructs the data type."""
        r, c = self._sizes[idx]
        if issubclass(
            self.type, ImageArray
        ):  # IF the underlying type is an ImageArray, then return as a view with extra metadata
            tmp = self._stack[:r, :c, idx].view(type=self.type)
        else:  # Otherwise it must be something with a data attribute
            tmp = self.type()  # pylint: disable=E1102
            tmp.data = self._stack[:r, :c, idx]
        tmp.metadata = deepcopy(self._common_metadata)
        tmp.metadata.update(self._metadata[self.__names__()[idx]])
        tmp.metadata["Scan #"] = self.scan_no
        tmp._fromstack = True
        return tmp

    def _load_files(self, root_name, regrid):
        """Build the image stack from a stack of files.

        Args:
            root_name (str):
                The scan prefix e.g. SC_###
            regrid (bool):
                Use the X and Y positions if available to regrid the data.
        """
        if not path.exists(path.join(self.directory, f"{root_name}-Parameters.txt")):
            return False
        self._load_parameters(root_name)
        for data in glob(path.join(self.directory, f"{root_name}*.asc")):
            if data.endswith("fwd.asc"):
                if "fwd" not in self.groups:
                    self.add_group("fwd")
                target = self.groups["fwd"]
            elif data.endswith("bwd.asc"):
                if "bwd" not in self.groups:
                    self.add_group("bwd")
                target = self.groups["bwd"]
            else:
                target = self
            target._load_asc(data)
        if regrid:
            if "fwd" in self.groups:
                self.groups["fwd"].regrid(in_place=True)
            if "bwd" in self.groups:
                self.groups["bwd"].regrid(in_place=True)
            if not self.groups:
                self.regrid(in_place=True)
        return self

    def _load_parameters(self, root_name):
        """Load the scan parameters text file.

        Args:
            root_name (str):
                The scan prefix e.g. SC_###

        Returns:
            self:
                The modififed scan stack.
        """
        filename = path.join(self.directory, f"{root_name}-Parameters.txt")
        with FileManager(filename, "r") as parameters:
            if not parameters.readline().startswith("Daisy Parameter Snapshot"):
                raise IOError("Parameters file exists but does not have correct header")
            for line in parameters:
                if not line.strip():
                    continue
                parts = [x.strip() for x in line.strip().split(":")]
                key = parts[0]
                value = ":".join(parts[1:])
                units = PARAM_RE.match(value)
                if units and units.groups()[1]:
                    key += f" [{units.groups()[1]}]"
                    value = units.groups()[0]
                self._common_metadata[key] = value
        return self

    def _load_asc(self, filename):
        """Load a single scan file from ascii data."""
        with FileManager(filename, "r") as data:
            if not data.readline().startswith("# Daisy frame view snapshot"):
                raise ValueError(f"{filename} lacked the correct header line")
            tmp = ImageFile()
            for line in data:
                if not line.startswith("# "):
                    break
                parts = [x.strip() for x in line[2:].strip().split(":")]
                key = parts[0]
                value = ":".join(parts[1:])
                units = PARAM_RE.match(value)
                if units and units.groups()[1]:
                    key += f" [{units.groups()[1]}]"
                    value = units.groups()[0]
                tmp.metadata[key] = value
        xpx = tmp["x-pixels"]
        ypx = tmp["y-pixels"]
        metadata = tmp.metadata
        tmp.image = genfromtxt(filename).reshape((xpx, ypx))
        tmp.metadata = metadata
        tmp.filename = tmp["display"]
        self.append(tmp)
        return self

    def _read_signal(self, g):
        """Read a signal array and return a member of the image stack."""
        if "signal" not in g:
            raise StonerLoadError(f"{g.name} does not have a signal dataset !")
        tmp = self.type()  # pylint: disable=E1102
        data = g["signal"]
        if prod(array(data.shape)) > 0:
            tmp.image = data[...]
        else:
            tmp.image = [[]]
        metadata = g.require_group("metadata")
        typehints = g.get("typehints", None)
        if not isinstance(typehints, h5py.Group):
            typehints = dict()
        else:
            typehints = typehints.attrs
        for i in sorted(metadata.attrs):
            v = metadata.attrs[i]
            t = typehints.get(i, "Detect")
            if isinstance(v, string_types) and t != "Detect":  # We have typehints and this looks like it got exported
                tmp.metadata[f"{i}{{{t}}}".strip()] = f"{v}".strip()
            else:
                tmp[i] = metadata.attrs[i]
        tmp.filename = path.basename(g.name)
        return tmp

    def regrid(self, **kargs):
        """Regrid the data sets based on PosX and PosY channels.

        Keyword Parameters:
            x_range, y_range (tuple of start, stop, points):
                Range of x-y co-rdinates to regrid the data to. Used as an argument to :py:func:`np.linspace` to
                generate the coordinate
                vector.
            in_place (bool):
                If True then replace the existing datasets with the regridded data, otherwise create a new copy
                of the scan object. Default is False.

        Returns:
            (AttocubeScan):
                Scan object with regridded data. May be the same as the source object if in_place is True.
        """
        if not kargs.get("in_place", False):
            new = self.clone
        else:
            new = self
        try:
            x = self["PosX"]
            y = self["PosY"]
        except KeyError:  # Can't get X and Y data
            return new

        xrange = kargs.pop("x_range", (x[:, 0].max(), x[:, -1].min(), x.shape[1]))
        yrange = kargs.pop("y_range", (y[0].max(), y[-1].min(), y.shape[0]))
        nx, ny = meshgrid(linspace(*xrange), linspace(*yrange))
        for data in self.channels:
            if "PosX" in data or "PosY" in data:
                continue
            z = self[data]
            nz = griddata((x.ravel(), y.ravel()), z.ravel(), (nx.ravel(), ny.ravel()), method="cubic").reshape(
                nx.shape
            )
            new[data].data = nz
        new["PosX"].data = nx
        new["PosY"].data = ny

        return new

    def level_image(self, method="plane", signal="Amp"):
        """Remove a background signla by fitting an appropriate function.

        Keyword Arguments:
            method (str or callable):
                Eirther the name of a fitting function in the global scope, or a callable. *plane* and *parabola*
                are already defined.
            signal (str):
                The name of the dataset to be flattened. Defaults to the Amplitude signal

        Returns:
            (AttocubeScan):
                The current scan object with the data modified.
        """
        if isinstance(method, string_types):
            method = globals()[method]
        if not callable(method):
            raise ValueError("Could not get a callable method to flatten the data")
        data = self[signal]
        ys, xs = data.shape
        X, Y = meshgrid(linspace(-1, 1, xs), linspace(-1, 1, ys))
        Z = data.data
        X = X.ravel()
        Y = Y.ravel()
        Z = Z.ravel()

        popt = curve_fit(method, (X, Y), Z)[0]

        nZ = method((X, Y), *popt)
        Z -= nZ

        data.data = Z.reshape(xs, ys)
        return self

    def to_hdf5(self, filename=None):
        """Save the AttocubeScan to an hdf5 file."""
        mode = ""
        if filename is None:
            filename = path.join(self.directory, f"SC_{self.scan_no:03d}.hdf5")
        if isinstance(filename, pathlib.PurePath):
            filename = str(filename)
        if filename is None or (isinstance(filename, bool) and not filename):  # now go and ask for one
            filename = self.__file_dialog("w")
        if isinstance(filename, string_types):
            mode = "r+" if path.exists(filename) else "w"
        self.filename = filename
        with HDFFileManager(self.filename, mode=mode) as f:
            f.attrs["type"] = type(self).__name__
            f.attrs["module"] = type(self).__module__
            f.attrs["scan_no"] = self.scan_no
            f.attrs["groups"] = list(self.groups.keys())
            f.attrs["channels"] = list(self.channels)
            if "common_metadata" in f.parent and "common_metadata" not in f:
                f["common_metadata"] = h5py.SoftLink(f.parent["common_metadata"].name)
                f["common_typehints"] = h5py.SoftLink(f.parent["common_typehints"].name)
            else:
                metadata = f.require_group("common_metadata")
                typehints = f.require_group("common_typehints")
                for k in self._common_metadata:
                    try:
                        typehints.attrs[k] = self._common_metadata._typehints[k]
                        metadata.attrs[k] = self._common_metadata[k]
                    except TypeError:
                        # We get this for trying to store a bad data type - fallback to metadata export to string
                        parts = self._common_metadata.export(k).split("=")
                        metadata.attrs[k] = "=".join(parts[1:])

            for g in self.groups:  # Recurse to save groups
                grp = f.require_group(g)
                self.groups[g].to_hdf5(grp)

            for ch in self.channels:
                signal = f.require_group(ch)
                data = self[ch]
                signal.require_dataset(
                    "signal",
                    data=data.data,
                    shape=data.shape,
                    dtype=data.dtype,
                    compression=self.compression,
                    compression_opts=self.compression_opts,
                )
                metadata = signal.require_group("metadata")
                typehints = signal.require_group("typehints")
                for k in [x for x in data.metadata if x not in self._common_metadata]:
                    try:
                        typehints.attrs[k] = data.metadata._typehints[k]
                        metadata.attrs[k] = data.metadata[k]
                    except TypeError:
                        # We get this for trying to store a bad data type - fallback to metadata export to string
                        parts = data.metadata.export(k).split("=")
                        metadata.attrs[k] = "=".join(parts[1:])
        return self

    @classmethod
    def read_hdf5(cls, filename, *args, **kargs):
        """Create a new instance from an hdf file."""
        self = cls(regrid=False)
        if filename is None or not filename:
            self.get_filename("r")
            filename = self.filename
        else:
            self.filename = filename
        with HDFFileManager(filename, "r") as f:
            self.scan_no = f.attrs["scan_no"]
            if "groups" in f.attrs:
                sub_grps = f.attrs["groups"]
            else:
                sub_grps = None
            if "channels" in f.attrs:
                channels = f.attrs["channels"]
            else:
                channels = []
            grps = list(f.keys())
            if "common_metadata" not in grps or "common_typehints" not in grps:
                _raise_error(f, message="Couldn;t find common metadata groups, something is not right here!")
            metadata = f["common_metadata"].attrs
            typehints = f["common_typehints"].attrs
            for i in sorted(metadata):
                v = metadata[i]
                t = typehints.get(i, "Detect")
                if (
                    isinstance(v, string_types) and t != "Detect"
                ):  # We have typehints and this looks like it got exported
                    self._common_metadata[f"{i}{{{t}}}".strip()] = f"{v}".strip()
                else:
                    self._common_metadata[i] = metadata[i]
            grps.remove("common_metadata")
            grps.remove("common_typehints")
            if sub_grps is None:
                sub_grps = grps
            for grp in sub_grps:
                if "type" in f[grp].attrs:
                    self.groups[grp] = cls.read_hdf5(f[grp], *args, **kargs)
                    continue
                g = f[grp]
                self.append(self._read_signal(g))
            for grp in channels:
                g = f[grp]
                self.append(self._read_signal(g))
        return self


class AttocubeScan(AttocubeScanMixin, ImageStack):
    """An ImageStack subclass that can load scans from the AttocubeScan SPM System.

    AttocubeScan represents a scan from an Attocube SPM system as a 3D stack of scan data with
    associated metadata. Indexing the AttocubeScan with either an integer or a partial match to one the
    signals saved in the scan will pull out that particular scan as a :py:class:`Stoner.Image.ImageFile`.

    If the scan was a dual pass scan with forwards and backwards data, then the root AttocubeScan will
    contain the common metadata derived from the Scan parameters and then two sub-stacks that represent
    the forwards ('fwd') and backwards ('bwd') scans.

    The AttocubeScan constructor will with take a *rrot name* of the scan - e.g. "SC_099" or alternatively
    a scan number integer. It will then look in the stack's directory for matching files and builds the scan stack
    from them. Currently, it uses the scan parameters.txt file and any ASCII stream files .asc files. 2D linescans
    are not currently supported or imported.

    The native file format for an AttocubeScan is an HDF5 file with a particilar structure. The stack is saved into
    an HDF5 group which then has a *type* and *module* attribute that specifies the class and module pf the Python
    object that created the group - sof for an AttocubeScan, the type attribute is *AttocubeScan*.

    There is a class method :py:meth:`AttocubeSca.read_hdf5` to read the stack from the HDSF format and an instance
    method :py:meth:`AttocubeScan.to_hdf` that will save to either a new or existing HDF file format.

    The class provides other methods to regrid and flatten images and may gain other capabilities in the future.

    Todo:
        Implement load and save to/from multipage TIFF files.

    Attrs:
        scan_no (int):
            The scan number as defined in the Attocube software.
        compression (str):
            The HDF5 compression algorithm to use when writing files
        compression_opts (int):
            The lelbel of compression to use (depends on compression algorithm)
    """



# -*- coding: utf-8 -*-
"""Test Stoner.Data as a sequence and mapping structure."""

import pytest
import sys
from os import path
import numpy as np
from numpy import all

from Stoner import Data, __home__

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
sys.path.insert(0, pth)
datadir = path.join(pth, "sample-data")
selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")


def test_indexing():
    # Check all the indexing possibilities
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    data = np.array(selfd.data)
    colname = selfd.column_headers[0]
    assert all(selfd.column(colname) == selfd[:, 0]), "Failed direct indexing versus column method"
    assert all(selfd[:, 0] == data[:, 0]), "Failed direct idnexing versusus direct array index"
    assert all(selfd[:, [0, 1]] == data), "Failed direct list indexing"
    assert all(selfd[::2, :] == data[::2]), "Failed slice indexing rows"
    assert all(selfd[colname] == data[:, 0]), "Failed direct indexing by column name"
    assert all(selfd[:, colname] == data[:, 0]), "Failed fallback indexing by column name"
    assert selfd[25, 1] == 645.0, "Failed direct single cell index"
    assert selfd[25, "Y-Data"] == 645.0, "Failoed single cell index direct"
    assert selfd["Y-Data", 25] == 645.0, "Failoed single cell fallback index order"
    selfd["X-Dat"] = [11, 12, 13, 14, 15]
    assert selfd["X-Dat", 2] == 13, "Failed indexing of metadata lists with tuple"
    assert selfd["X-Dat"][2] == 13, "Failed indexing of metadata lists with double indices"
    d = Data(np.ones((10, 10)))
    d[0, 0] = 5  # Index by tuple into data
    d["Column_1", 0] = 6  # Index by column name, row into data
    d[0, "Column_2"] = 7  # Index by row, column name into data
    d["Column_3"] = [1, 2, 3, 4]  # Create a metadata
    d["Column_3", 2] = 2  # Index existing metadata via tuple
    d.metadata[0, 5] = 10
    d[0, 5] = 12  # Even if tuple, index metadata if already existing.
    assert np.all(
        d[0] == np.array([5, 6, 7, 1, 1, 1, 1, 1, 1, 1])
    ), f"setitem on Data to index into Data.data failed.\n{d[0]}"
    assert d.metadata["Column_3"] == [1, 2, 2, 4], "Tuple indexing into metadata Failed."
    assert d.metadata[0, 5] == 12, "Indexing of pre-existing metadta keys rather than Data./data failed."
    assert d.metadata[1] == [1, 2, 2, 4], "Indexing metadata by integer failed."
    with pytest.raises(KeyError):
        selfd["Bad key"]
    selfd["tmp"] = np.linspace(1, 9, 9).reshape((3, 3))
    assert all(selfd["tmp", 1] == np.array([4, 5, 6]))
    selfd["tmp", 1, 1] = -5
    assert all(selfd["tmp", 1] == np.array([4, -5, 6]))
    assert all(selfd["X-Data"] == selfd.x)


def test_metadata():
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    selfd3 = Data(path.join(__home__, "..", "sample-data", "New-XRay-Data.dql"))
    selfd["Test"] = "This is a test"
    selfd["Int"] = 1
    selfd["Float"] = 1.0
    keys, values = zip(*selfd.items())
    assert tuple(selfd.keys()) == keys, "Keys from items() not equal to keys from keys()"
    assert tuple(selfd.values()) == values, "Values from items() not equal to values from values()"
    assert selfd["Int"] == 1
    assert selfd["Float"] == 1.0
    assert selfd["Test"] == selfd.metadata["Test"]
    assert selfd.metadata._typehints["Int"] == "I32"
    assert len(selfd.dir()) == 6, f"Failed meta data directory listing ({selfd.dir()})"
    assert len(selfd3["Temperature"]) == 7, "Regular expression metadata search failed"


def test_len():
    # Check that length of the column is the same as length of the data
    selfd = Data(path.join(path.dirname(__file__), "..", "CoreTest.dat"), setas="xy")
    assert len(Data()) == 0, "Empty DataFile not length zero"
    assert len(selfd.column(0)) == len(selfd), "Column 0 length not equal to DataFile length"
    assert len(selfd) == selfd.data.shape[0], "DataFile length not equal to data.shape[0]"
    # Check that self.column_headers returns the right length
    assert len(selfd.column_headers) == selfd.data.shape[1], "Length of column_headers not equal to data.shape[1]"


if __name__ == "__main__":
    pytest.main(["--pdb", __file__])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Filtering and smoothing functions for analysis code."""

__all__ = ["FilteringOpsMixin"]

from copy import deepcopy as copy
from warnings import warn

import numpy as np
from numpy import ma
from scipy.interpolate import interp1d, UnivariateSpline
from scipy.signal import get_window, convolve, savgol_filter

from Stoner.tools import isiterable, isnone
from Stoner.compat import int_types, string_types, get_func_params

from .utils import outlier as _outlier, _twoD_fit, GetAffineTransform


class FilteringOpsMixin:
    """Provide additional filtering sndsmoothing methods to :py:class:`Stoner.Data`."""

    def SG_Filter(
        self, col=None, xcol=None, points=15, poly=1, order=0, pad=True, result=None, replace=False, header=None
    ):
        """Implement a Savitsky-Golay filtering of data for smoothing and differentiating data.

        Args:
            col (column index):
                Column of Data to be filtered. if None, first y-column in setas is filtered.
            points (int):
                Number of data points to use in the filtering window. Should be an odd number > poly+1 (default 15)

        Keyword Arguments:
            xcol (coilumn index):
                If *order*>1 then can be used to specify an x-column to differentiate with respect to.
            poly (int):
                Order of polynomial to fit to the data. Must be equal or greater than order (default 1)
            order (int):
                Order of differentiation to carry out. Default=0 meaning smooth the data only.
            pad (bool or float):
                Pad the start and end of the array with the mean value (True, default) or specired value (float) or
                leave as is.
            result (None,True, or column_index):
                If not None, column index to insert new data, or True to append as last column
            header (string or None):
                Header for new column if result is not None. If header is Nne, a suitable column header is generated.

        Returns:
            (numpy array or self):
                If result is None, a numpy array representing the smoothed or differentiated data is returned.
                Otherwise, a copy of the modified AnalysisMixin object is returned.

        Notes:
            If col is not specified or is None then the :py:attr:`DataFile.setas` column assignments are used
            to set an x and y column. If col is a tuple, then it is assumed to specify and x-column and y-column
            for differentiating data. This is now a pass through to :py:func:`scipy.signal.savgol_filter`

            Padding can help stop wildly wrong artefacts in the data at the start and enf of the data, particularly
            when the differential order is >1.

        See Also:
            User guide section :ref:`smoothing_guide`
        """
        from Stoner.Util import ordinal

        points = int(points)
        if points % 2 == 0:  # Ensure window length is odd
            points += 1

        _ = self._col_args(scalar=False, ycol=col, xcol=xcol)

        if _.xcol is not None:
            if not isinstance(_.xcol, list):
                col = _.ycol + [_.xcol]
            else:
                col = _.ycol + _.xcol
            data = self.column(col).T
        else:
            col = _.ycol
            data = self.column(list(col)).T
            data = np.row_stack((data, np.arange(data.shape[1])))

        ddata = savgol_filter(data, window_length=points, polyorder=poly, deriv=order, mode="interp")
        if isinstance(pad, bool) and pad:
            offset = int(np.ceil(points * (order + 1) ** 2 / 8))
            padv = np.mean(ddata[:, offset:-offset], axis=1)
            pad = np.ones((ddata.shape[0], offset))
            for ix, v in enumerate(padv):
                pad[ix] *= v
        elif isinstance(pad, float):
            offset = int(np.ceil(points / 2))
            pad = np.ones((ddata.shape[0], offset)) * pad

        if np.all(pad) and offset > 0:
            ddata[:, :offset] = pad
            ddata[:, -offset:] = pad
        if order >= 1:
            r = ddata[:-1] / ddata[-1]
        else:
            r = ddata

        if result is not None:
            if not isinstance(header, string_types):
                header = []
                for column in col[:-1]:
                    header.append(f"{self.column_headers[column]} after {ordinal(order)} order Savitsky-Golay Filter")
            else:
                header = [header] * (len(col) - 1)
            if r.shape[0] > len(header):
                iterdata = r[: len(header)]
            else:
                iterdata = r
            for column, head in zip(iterdata, header):
                self.add_column(column.ravel(), header=head, index=result, replace=replace)
            return self
        return r

    def bin(self, xcol=None, ycol=None, bins=0.03, mode="log", clone=True, **kargs):
        """Bin x-y data into new values of x with an error bar.

        Args:
            xcol (index):
                Index of column of data with X values
            ycol (index):
                Index of column of data with Y values
            bins (int, float or 1d array):
                Number of bins (if integer) or size of bins (if float), or bin edges (if array)
            mode (string):
                "log" or "lin" for logarithmic or linear binning

        Keyword Arguments:
            yerr (index):
                Column with y-error data if present.
            bin_start (float):
                Manually override the minimum bin value
            bin_stop (float):
                Manually override the maximum bin value
            clone (bool):
                Return a clone of the current AnalysisMixin with binned data (True)
                or just the numbers (False).

        Returns:
            (:py:class:`Stoner.Data` or tuple of 4 array-like):
                Either a clone of the current data set with the new binned data or
                tuple of (bin centres, bin values, bin errors, number points/bin),
                depending on the *clone* parameter.

        Note:
            Algorithm inspired by MatLab code wbin,    Copyright (c) 2012:
            Michael Lindholm Nielsen


        See Also:
            User Guide section :ref:`binning_guide`
        """

        if None in (xcol, ycol):
            cols = self.setas._get_cols()
            if xcol is None:
                xcol = cols["xcol"]
            if ycol is None:
                ycol = cols["ycol"]
        yerr = kargs.pop("yerr", cols["yerr"] if cols["has_yerr"] else None)

        bin_left, bin_right, bin_centres = self.make_bins(xcol, bins, mode, **kargs)

        ycol = self.find_col(ycol)
        if yerr is not None:
            yerr = self.find_col(yerr)

        ybin = np.zeros((len(bin_left), len(ycol)))
        ebin = np.zeros((len(bin_left), len(ycol)))
        nbins = np.zeros((len(bin_left), len(ycol)))
        xcol = self.find_col(xcol)
        i = 0

        for limits in zip(bin_left, bin_right):
            data = self.search(xcol, limits)
            if len(data) > 1:
                ok = np.logical_not(np.isnan(data.y))
                data = data[ok]
            elif len(data) == 0 or (len(data) == 1 and np.isnan(data.y)):
                shape = list(data.shape)
                shape[0] = 0
                data = np.zeros(shape)
            if yerr is not None:
                w = 1.0 / data[:, yerr] ** 2
                W = np.sum(w, axis=0)
                if data.shape[0] > 3:
                    e = max(np.std(data[:, ycol], axis=0) / np.sqrt(data.shape[0]), (1.0 / np.sqrt(W)) / data.shape[0])
                else:
                    e = 1.0 / np.sqrt(np.where(W > 0, W, np.nan))
            else:
                w = np.ones((data.shape[0], len(ycol)))
                W = data.shape[0]
                if data[:, ycol].size > 1:
                    e = np.std(data[:, ycol], axis=0) / np.sqrt(W)
                else:
                    e = np.nan
            if data.shape[0] == 0 and self.debug:
                warn(f"Empty bin at {limits}")
            y = np.sum(data[:, ycol] * (w / W), axis=0)
            ybin[i, :] = y
            ebin[i, :] = e
            nbins[i, :] = data.shape[0]
            i += 1
        if clone:
            ret = self.clone
            ret.data = np.atleast_2d(bin_centres).T
            ret.column_headers = [self.column_headers[xcol]]
            ret.setas = ["x"]
            for i in range(ybin.shape[1]):
                head = str(self.column_headers[ycol[i]])

                ret.add_column(ybin[:, i], header=head)
                ret.add_column(ebin[:, i], header=f"d{head}")
                ret.add_column(nbins[:, i], header=f"#/bin {head}")
                s = list(ret.setas)
                s[-3:] = ["y", "e", "."]
                ret.setas = s
        else:
            ret = (bin_centres, ybin, ebin, nbins)
        return ret

    def extrapolate(self, new_x, xcol=None, ycol=None, yerr=None, overlap=20, kind="linear", errors=None):
        """Extrapolate data based on local fit to x,y data.

        Args:
            new_x (float or array):
                New values of x data.

        Keyword Arguments:
            xcol (column index, None):
                column containing x-data or None to use setas attribute
            ycol (column index(es) or None):
                column(s) containing the y-data or None to use setas attribute.
            yerr (column index(es) or None):
                y error data column or None to use setas attribute
            overlap (float or int):
                range of x-data used for the local fit for extrapolating. If int then overlap number of
                points is used, if float then that range x-axis space is used.
            kind (str or callable):
                Determines local fitting function. If string should be "linear", "quadratic" or "cubic" if
                callable, then represents a function to be fitted to the data.
            errors (callable or None):
                If *kind* is a callable function, then errs must be defined and must also be a callable function.

        Returns:
            (array):
                Extrapolated values.

        Note:
            If the new_x values lie outside the span of the x-data, then the nearest *overlap* portion of the data
            is used to estimate the values. If the new_x values are within the span of the x-data then the portion
            of the data centred about the point and overlap points long will be used to interpolate a value.

            If *kind* is callable, it should take x values in the first parameter and free fitting parameters as
            the other parameters (i.e. as with :py:meth:`AnalysisMixin.curve_fit`).
        """
        _ = self._col_args(xcol=xcol, ycol=ycol, yerr=yerr, scalar=False)
        kinds = {
            "linear": lambda x, m, c: m * x + c,
            "quadratic": lambda x, a, b, c: a * x**2 + b * x + c,
            "cubic": lambda x, a, b, c, d: a * x**3 + b * x**2 + c * x + d,
        }
        errs = {
            "linear": lambda x, me, ce, popt: np.sqrt((me * x) ** 2 + ce**2),
            "quadratic": lambda x, ae, be, ce, popt: np.sqrt((2 * x**2 * ae) ** 2 + (x * be) ** 2 + ce**2),
            "cubic": lambda x, ae, be, ce, de, popt: np.sqrt(
                (3 * ae * x**3) ** 2 + (2 * x**2 * be) ** 2 + (x * ce) ** 2 + de**2
            ),
        }

        if callable(kind):
            kindf = kind
            if not callable(errors):
                raise TypeError("If kind is a callable, then errs must be defined and be callable as well")
            errsf = errors
        elif kind in kinds:
            kindf = kinds[kind]
            errsf = errs[kind]
        else:
            raise RuntimeError(f"Failed to recognise extrpolation function '{kind}'")
        scalar_x = not isiterable(new_x)
        if scalar_x:
            new_x = [new_x]
        if isinstance(new_x, ma.MaskedArray):
            new_x = new_x.compressed
        results = np.zeros((len(new_x), 2 * len(_.ycol)))
        work = self.clone
        for ix, x in enumerate(new_x):
            r = self.closest(x, xcol=_.xcol)
            if isinstance(overlap, int):
                if (r.i - overlap / 2) < 0:
                    ll = 0
                    hl = min(len(self), overlap)
                elif (r.i + overlap / 2) > len(self):
                    hl = len(self)
                    ll = max(hl - overlap, 0)
                else:
                    ll = r.i - overlap / 2
                    hl = r.i + overlap / 2
                bounds = {"_i__between": (ll, hl)}
                mid_x = (self[ll, _.xcol] + self[hl - 1, _.xcol]) / 2.0
            elif isinstance(overlap, float):
                if (r[_.xcol] - overlap / 2) < self.min(_.xcol)[0]:
                    ll = self.min(_.xcol)[0]
                    hl = ll + overlap
                elif (r[_.xcol] + overlap / 2) > self.max(_.xcol)[0]:
                    hl = self.max(_.xcol)[0]
                    ll = hl - overlap
                else:
                    ll = r[_.xcol] - overlap / 2
                    hl = r[_.xcol] + overlap / 2
                bounds = {f"{self.column_headers[_.xcol]}__between": (ll, hl)}
                mid_x = (ll + hl) / 2.0
            pointdata = work.select(**bounds)
            pointdata.data[:, _.xcol] = pointdata.column(_.xcol) - mid_x
            ret = pointdata.curve_fit(kindf, _.xcol, _.ycol, sigma=_.yerr, absolute_sigma=True)
            if isinstance(ret, tuple):
                ret = [ret]
            for iy, rt in enumerate(ret):
                popt, pcov = rt
                perr = np.sqrt(np.diag(pcov))
                results[ix, 2 * iy] = kindf(x - mid_x, *popt)
                results[ix, 2 * iy + 1] = errsf(x - mid_x, *perr, popt)
        if scalar_x:
            results = results[0]
        return results

    def interpolate(self, newX, kind="linear", xcol=None, replace=False):
        """Interpolate a dataset to get a new set of values for a given set of x data.

        Args:
            ewX (1D array or None):
                Row indices or X column values to interpolate with. If None, then the
                :py:meth:`AnalysisMixin.interpolate` returns an interpolation function. Unlike the raw interpolation
                function from scipy, this interpolation function will work with MaskedArrays by compressing them
                first.

        Keyword Arguments:
            kind (string):
                Type of interpolation function to use - does a pass through from numpy. Default is linear.
            xcol (index or None):
                Column index or label that contains the data to use with newX to determine which rows to return.
                Defaults to None.
            replace (bool):
                If true, then the current AnalysisMixin's data is replaced with the  newly interpolated data and the
                current AnalysisMixin is returned.

        Returns:
            (2D numpy array):
                Section of the current object's data if replace is False(default) or the modofied AnalysisMixin if
                replace is true.

        Note:
            Returns complete rows of data corresponding to the indices given in newX. if xcol is None, then newX is
            interpreted as (fractional) row indices. Otherwise, the column specified in xcol is thresholded with the
            values given in newX and the resultant row indices used to return the data.

            If the positional argument, newX is None, then the return value is an interpolation function. This
            interpolation function takes one argument - if *xcol* was None, this argument is interpreted as
            array indices, but if *xcol* was specified, then this argument is interpreted as an array of xvalues.
        """
        DataArray = type(self.data)  # pylint: disable=E0203
        lines = np.shape(self.data)[0]  # pylint: disable=E0203
        index = np.arange(lines)
        if xcol is None:
            xcol = self.setas._get_cols("xcol")
        elif isinstance(xcol, bool) and not xcol:
            xcol = None

        if isinstance(newX, ma.MaskedArray):
            newX = newX.compressed()

        if xcol is not None and newX is not None:  # We need to convert newX to row indices
            xfunc = interp1d(self.column(xcol), index, kind, 0)  # xfunc(x) returns partial index
            newX = xfunc(newX)
        inter = interp1d(index, self.data, kind, 0)  # pylint: disable=E0203

        if newX is None:  # Ok, we're going to return an interpolation function

            def wrapper(newX):
                """Wrap the interpolation function."""
                if isinstance(newX, ma.MaskedArray):
                    newX = newX.compressed()
                else:
                    newX = np.array(newX)
                if xcol is not None and newX is not None:  # We need to convert newX to row indices
                    xfunc = interp1d(self.column(xcol), index, kind, 0)  # xfunc(x) returns partial index
                    newX = xfunc(newX)
                return inter(newX)

            return wrapper

        if replace:
            self.data = inter(newX)
            ret = self
        else:
            ret = DataArray(inter(newX), isrow=True)
            ret.setas = self.setas.clone
        return ret

    def make_bins(self, xcol, bins, mode="lin", **kargs):
        """Generate bin boundaries and centres along an axis.

        Args:
            xcol (index):
                Column of data with X values
            bins (1d_)array or int or float):
                Number of bins (int) or width of bins (if float)
            mode (string):
                "lin" for linear binning, "log" for logarithmic binning.

        Keyword Arguments:
            bin_start (float):
                Override minimum bin value
            bin_stop (float):
                Override the maximum bin value

        Returns:
            (tuple of 4 arrays):
                bin_start,bin_stop,bin_centres (1D arrays): The locations of the bin
                boundaries and centres for each bin.
        """
        xmin = kargs.pop("bin_start", (self // xcol).min())
        xmax = kargs.pop("bin_sop", (self // xcol).max())

        if isinstance(bins, int):  # Given a number of bins
            if mode.lower().startswith("lin"):
                bin_width = float(xmax - xmin) / bins
                bin_start = np.linspace(xmin, xmax - bin_width, bins)
                bin_stop = np.linspace(xmin + bin_width, xmax, bins)
                bin_centres = (bin_start + bin_stop) / 2.0
            elif mode.lower().startswith("log"):
                xminl = np.log(xmin)
                xmaxl = np.log(xmax)
                bin_width = float(xmaxl - xminl) / bins
                bin_start = np.linspace(xminl, xmaxl - bin_width, bins)
                bin_stop = np.linspace(xminl + bin_width, xmaxl, bins)
                bin_centres = (bin_start + bin_stop) / 2.0
                bin_start = np.exp(bin_start)
                bin_stop = np.exp(bin_stop)
                bin_centres = np.exp(bin_centres)
            else:
                raise ValueError(f"mode should be either lin(ear) or log(arthimitc) not {mode}")
        elif isinstance(bins, float):  # Given a bin with as a flot
            if mode.lower().startswith("lin"):
                bin_width = bins
                bins = int(np.ceil(abs(float(xmax - xmin) / bins)))
                bin_start = np.linspace(xmin, xmax - bin_width, bins)
                bin_stop = np.linspace(xmin + bin_width, xmax, bins)
                bin_centres = (bin_start + bin_stop) / 2.0
            elif mode.lower().startswith("log"):
                if not 0.0 < bins <= 1.0:
                    raise ValueError("Bin width must be between 0 and 1 for log binning")
                if xmin <= 0:
                    raise ValueError("The start of the binning must be a positive value in log mode.")
                xp = xmin
                splits = []
                centers = []
                while xp < xmax:
                    splits.append(xp)
                    centers.append(xp * (1 + bins / 2))
                    xp = xp * (1 + bins)
                splits.append(xmax)
                bin_start = np.array(splits[:-1])
                bin_stop = np.array(splits[1:])
                bin_centres = np.array(centers)
            else:
                raise ValueError(f"mode should be either lin(ear) or log(arthimitc) not {mode}")
        elif isinstance(bins, np.ndarray) and bins.ndim == 1:  # Yser provided manuals bins
            bin_start = bins[:-1]
            bin_stop = bins[1:]
            if mode.lower().startswith("lin"):
                bin_centres = (bin_start + bin_stop) / 2.0
            elif mode.lower().startswith("log"):
                bin_start = np.where(bin_start <= 0, 1e-9, bin_start)
                bin_centres = np.exp(np.log(bin_start) + np.log(bin_stop) / 2.0)
            else:
                raise ValueError(f"mode should be either lin(ear) or log(arthimitc) not {mode}")
        else:
            raise TypeError(f"bins must be either an integer or a float, not a {type(bins)}")
        if len(bin_start) > len(self):
            raise ValueError("Attempting to bin into more bins than there is data.")
        return bin_start, bin_stop, bin_centres

    def outlier_detection(
        self, column=None, window=7, shape="boxcar", certainty=3.0, action="mask", width=1, func=None, **kargs
    ):
        """Detect outliers in a column of data.

        Args:
            column(column index):
                specifying column for outlier detection. If not set,
                defaults to the current y set column.

        Keyword Arguments:
            window(int):
                data window for anomaly detection
            shape(str):
                The name of a :py:mod:`scipy.signal` windowing function to use when averaging the data.
                Defaults to 'boxcar' for a flat average.
            certainty(float):
                eg 3 detects data 3 standard deviations from average
            action(str or callable):
                what to do with outlying points, options are
                * 'mask' outlier points are masked (default)
                * 'mask row' outlier rows are masked
                * 'delete'  outlier rows are deleted
                * callable  the value of the action keyword is called with the outlier row
                * anything else defaults to do nothing.

            width(odd integer):
                Number of rows that an outliing spike could occupy. Defaults to 1.
            func (callable):
                A function that determines if the current row is an outlier.
            action_args (tuple):
                if *action* is callable, then action_args can be used to pass extra arguments to the action callable
            action_kargs (dict):
                If *action* is callable, then action_kargs can be used to pass extra keyword arguments to the action
                callable.

        Returns:
            (:py:class:`Stoner.Data`):
                The newly modified Data object.

        outlier_detection will add row numbers of detected outliers to the metadata
        of d, also will perform action depending on request eg 'mask', 'delete'
        (any other action defaults to doing nothing).

        The detection looks at a window of the data, takes the average and looks
        to see if the current data point falls certainty * std deviations away from
        data average.

        The outlier detection function has the signatrure::

            def outlier(row,column,window,certainty,**kargs)
                #code
                return True # or False

        All extra keyword arguments are passed to the outlier detector.

        IF *action* is a callable function then it should take the form of::

            def action(i,column, data, *action_args, **action_kargs):
                pass

        where *i* is the number of the outlier row, *column* the same value as above
        and *data* is the complete set of data.

        In all cases the indices of the outlier rows are added to the ;outlier' metadata.

        Example:
            .. plot:: samples/outlier.py
                :include-source:
                :outname: outlier
        """
        if func is None:
            func = _outlier

        if action not in ["delete", "mask", "mask row"] and not callable(action):
            raise ValueError(f"Do'n know what to do with action={action}")
        _ = self._col_args(scalar=True, ycol=column, **kargs)
        column = _.ycol
        params = get_func_params(func)
        for p in params:
            if p in _:
                kargs[p] = _[p]
        kargs.setdefault("ycol", column)
        if not callable(action) and ("action_args" in kargs or "acation_kargs" in kargs):
            raise SyntaxError("Can only have action_args and action_kargs keywords in action is callable")
        action_args = kargs.pop("action_args", ())
        action_kargs = kargs.pop("action_kargs", {})
        kargs["shape"] = shape
        for k in list(kargs.keys()):
            if k not in params:
                kargs.pop(k)
        index = np.zeros(len(self), dtype=bool)
        for i, t in enumerate(self.rolling_window(window, wrap=False, exclude_centre=width)):
            index[i] = func(self.data[i], t, metric=certainty, **kargs)
        self["outliers"] = np.arange(len(self))[index]  # add outlier indices to metadata
        if action == "mask" or action == "mask row":
            if action == "mask":
                self.mask[index, column] = True
            else:
                self.mask[index, :] = True
        elif action == "delete":
            self.data = self.data[~index]
        elif callable(action):  # this will call the action function with each row in turn from back to start
            for i in np.arange(len(self))[index][::-1]:
                action(i, column, self.data, *action_args, **action_kargs)
        return self

    def scale(self, other, xcol=None, ycol=None, **kargs):
        """Scale the x and y data in this DataFile to match the x and y data in another DataFile.

        Args:
            other (DataFile):
                The other instance of a datafile to match to

        Keyword Arguments:
            xcol (column index):
                Column with x points in it, default to None to use setas attribute value
            ycol (column index):
                Column with ypoints in it, default to None to use setas attribute value
            xmode ('affine', 'linear','scale','offset'):
                How to manipulate the x-data to match up
            ymode ('linear','scale','offset'):
                How to manipulate the y-data to match up.
            bounds (callable):
                Used to identiyf the set of (x,y) points to be used for scaling. Defaults to the whole data set if
                not speicifed.
            otherbounds (callable):
                Used to detemrine the set of (x,y) points in the other data file. Defaults to bounds if not given.
            use_estimate (bool or 3x2 array):
                Specifies whether to estimate an initial transformation value or to use the provided one, or
                start with an identity transformation.
            replace (bool):
                Whether to map the x,y data to the new coordinates and return a copy of this AnalysisMixin (true)
                or to just return the results of the scaling.
            headers (2-element list or tuple of strings):
                new column headers to use if replace is True.

        Returns:
            (various):
                Either a copy of the :py:class:Stoner.Data` modified so that the x and y columns match *other*
                if *replace* is True, or *opt_trans*,*trans_err*,*new_xy_data*. Where *opt_trans* is the optimum
                affine transformation, *trans_err* is a matrix giving the standard error in the transformation
                matrix components and  *new_xy_data* is an (n x 2) array of the transformed data.

        Example:
            .. plot:: samples/scale_curves.py
                :include-source:
                :outname: scale
        """
        _ = self._col_args(xcol=xcol, ycol=ycol)
        #
        # Sort out keyword srguments
        #
        bounds = kargs.pop("bounds", lambda x, r: True)
        otherbounds = kargs.pop("otherbounds", bounds)
        replace = kargs.pop("replace", True)
        headers = kargs.pop("headers", None)
        xmode = kargs.pop("xmode", "linear")
        ymode = kargs.pop("ymode", "linear")
        use_estimate = kargs.pop("use_estimate", False)

        # Get our working data from this DataFile and remove masked rows

        working = self.search(_.xcol, bounds)
        working = ma.mask_rowcols(working, axis=0)
        xdat = working[:, self.find_col(_.xcol)]
        ydat = working[:, self.find_col(_.ycol)]

        # Get data from the other. If it is already an ndarray, check size and dimensions

        if isinstance(other, self._baseclass):
            working2 = other.search(_.xcol, otherbounds)
            working2 = ma.mask_rowcols(working2, axis=0)
            xdat2 = working2[:, other.find_col(_.xcol)]
            ydat2 = working2[:, other.find_col(_.ycol)]
            if len(xdat2) != len(xdat):
                raise RuntimeError(f"Data lengths don't match {len(xdat)}!={len(xdat2)}")
        elif isinstance(other, np.ndarray):
            if other.ndim == 1:
                other = np.atleast_2d(other).T
            if other.shape[0] != len(xdat) or not 1 <= other.shape[1] <= 2:
                raise RuntimeError(
                    (
                        "If other is a numpy array it must be the same length as the number of points to match "
                        + "to and 1 or 2 columns. (other shape={})"
                    ).format(other.shape)
                )
            if other.shape[1] == 1:
                xdat2 = xdat
                ydat2 = other[:, 0]
            else:
                xdat2 = other[:, 0]
                ydat2 = other[:, 1]
        else:
            raise RuntimeError(f"other should be either a numpy array or subclass of DataFile, not a {type(other)}")

        # Need two nx2 arrays of points now

        xy1 = np.column_stack((xdat, ydat))
        xy2 = np.column_stack((xdat2, ydat2))

        # We're going to use three points to get an estimate for the affine transform to apply

        if isinstance(use_estimate, bool) and use_estimate:
            mid = int(len(xdat) / 2)
            try:  # may go wrong if three points are co-linear
                m0 = GetAffineTransform(xy1[[0, mid, -1], :], xy2[[0, mid, -1], :])
            except (RuntimeError, np.linalg.LinAlgError):  # So use an idnetify transformation instead
                m0 = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        elif isinstance(use_estimate, np.ndarray) and use_estimate.shape == (
            2,
            3,
        ):  # use_estimate is an initial value transformation
            m0 = use_estimate
        else:  # Don't try to be clever
            m0 = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        popt, perr, trans = _twoD_fit(xy1, xy2, xmode=xmode, ymode=ymode, m0=m0)
        data = self.data[:, [_.xcol, _.ycol]]
        new_data = trans(data)
        if replace:  # In place scaling, replace and return self
            self.metadata["Transform"] = popt
            self.metadata["Transform Err"] = perr
            self.data[:, _.xcol] = new_data[:, 0]
            self.data[:, _.ycol] = new_data[:, 1]
            if headers:
                self.column_headers[_.xcol] = headers[0]
                self.column_headers[_.ycol] = headers[1]
            ret = self
        else:  # Return results but don't change self.
            ret = popt, perr, new_data
        return ret

    def smooth(self, window="boxcar", xcol=None, ycol=None, size=None, **kargs):
        """Smooth data by convoluting with a window.

        Args:
            window (string or tuple):
                Defines the window type to use by passing to :py:func:`scipy.signal.get_window`.

        Keyword Arguments:
            xcol(column index or None):
                Data to use as x data if needed to define a window. If None, use :py:attr:`Stoner.Core.DataFile.setas`
            ycvol (column index or None):
                Data to be smoothed
            size (int or float):
                If int, then the number of points to use in the smoothing window. If float, then the size in x-data
                to be used.
            result (bool or column index):
                Whether to add the smoothed data to the dataset and if so where.
            replace (bool):
                Replace the exiting data or insert as a new column.
            header (string):
                New column header for the new data.

        Returns:
            (self or array):
                If result is False, then the return value will be a copy of the smoothed data, otherwise the return
                value is a copy of the AnalysisMixin object with the smoothed data added,

        Notes:
            If size is float, then it is necessary to map the X-data to a number of rows and to ensure that the data
            is evenly spaced in x. To do this, the number of rows in the window is found by dividing the span in x
            by the size and multiplying by the total lenfth. Then the data is interpolated to a new set of evenly
            space X over the same range, smoothed and then interpoalted back to the original x values.
        """
        _ = self._col_args(xcol=xcol, ycol=ycol)
        replace = kargs.pop("replace", True)
        result = kargs.pop("result", True)  # overwrite existing y column data
        header = kargs.pop("header", self.column_headers[_.ycol])

        # Sort out window size
        if isinstance(size, float):
            interp_data = True
            xl, xh = self.span(_.xcol)
            size = int(np.ceil((size / (xh - xl)) * len(self)))
            nx = np.linspace(xl, xh, len(self))
            data = self.interpolate(nx, kind="linear", xcol=_.xcol, replace=False)
            self["Smoothing window size"] = size
        elif isinstance(size, int_types):
            data = copy(self.data)
            interp_data = False
        else:
            raise ValueError(f"size should either be a float or integer, not a {type(size)}")

        window = get_window(window, size)
        # Handle multiple or single y columns
        if not isiterable(_.ycol):
            _.ycol = [_.ycol]

        # Do the convolution itself
        for yc in _.ycol:
            data[:, yc] = convolve(data[:, yc], window, mode="same") / size

        # Reinterpolate the smoothed data back if necessary
        if interp_data:
            nx = self.data[:, _.xcol]
            tmp = self.clone
            tmp.data = data
            data = tmp.interpolate(nx, kind="linear", xcol=_.xcol, replace=False)

        # Fix return value
        if isinstance(result, bool) and not result:
            return data[:, _.ycol]
        for yc in _.ycol:
            self.add_column(data[:, yc], header=header, index=result, replace=replace)
        return self

    def spline(self, xcol=None, ycol=None, sigma=None, **kargs):
        """Construct a spline through x and y data and replace, add new data or return spline function.

        Keyword Arguments:
            xcol (column index):
                Column with x data or if None, use setas attribute.
            ycol (column index):
                Column with y data or if None, use the setas attribute
            sigma (column index, or array of data):
                Column with weights, or if None use the 1/yerr column.
            replace (Boolean or column index or None):
                If True then the y-column data is repalced, if a column index then the
                new data is added after the specified index, if False then the new y-data is returned and if None,
                then spline object is returned.
            header (string):
                If *replace* is True or a column index then use this string as the new column header.
            order (int):
                The order of spline to use (1-5)
            smoothing (float or None):
                The smoothing factor to use when fitting the spline. A value of zero will create an
                interpolating spline.
            bbox (tuple of length 2):
                Bounding box for the spline - defaults to range of x values
            ext (int or str):
                How to extrapolate, default is "extrapolate", but can also be "raise","zeros" or "const".

        Returns:
            (various):
                Depending on the value of *replace*, returns a copy of the AnalysisMixin, a 1D numpy array of
                data or an :[y:class:`scipy.interpolate.UniverateSpline` object.

        This is really just a pass through to the scipy.interpolate.UnivariateSpline function. Also used in the
        extrapolate function.
        """
        _ = self._col_args(xcol=xcol, ycol=ycol)
        if sigma is None and (isnone(_.yerr) or _.yerr):
            if not isnone(_.yerr):
                sigma = 1.0 / (self // _.yerr)
            else:
                sigma = np.ones(len(self))
        replace = kargs.pop("replace", True)
        result = kargs.pop("result", True)  # overwrite existing y column data
        header = kargs.pop("header", self.column_headers[_.ycol])
        k = kargs.pop("order", 3)
        s = kargs.pop("smoothing", None)
        bbox = kargs.pop("bbox", [None] * 2)
        ext = kargs.pop("ext", "extrapolate")
        x = self // _.xcol
        y = self // (_.ycol)
        spline = UnivariateSpline(x, y, w=sigma, bbox=bbox, k=k, s=s, ext=ext)
        new_y = spline(x)

        if header is None:
            header = self.column_headers[_.ycol]

        if not (result is None or (isinstance(result, bool) and not result)):
            self.add_column(new_y, header, index=result, replace=replace)
            return self
        if result is None:
            return new_y
        return spline



"""Detect peaks in a dataset."""

# pylint: disable=invalid-name,unsubscriptable-object
from matplotlib.cm import jet
from numpy import linspace

from Stoner import Data

d = Data("../../sample-data/New-XRay-Data.dql")
e = d.clone
d.plot()
d.peaks(width=0.08, poly=4, significance=100, modify=True)
d.labels[1] = "Peaks"
d.plot(fmt="ro")
e.find_peaks(modify=True, prominence=80, width=0.01)
e.labels[1] = "Find_Peaks"
e.plot(figure=d.fig, fmt="bx")
d.title = "Finding peaks with peaks() and find_peaks()"

# Use the metadata that find_peaks produces
colors = jet(linspace(0, 1, len(e)), alpha=0.5)
for l, r, c in zip(e["left_ips"], e["right_ips"], colors):
    ax = d.axes[d.ax]
    ax.axvspan(l, r, facecolor=c)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""":py:class:`lmfit.Model` model classes and functions for various magnetism and magnetic materials models."""
# pylint: disable=invalid-name
__all__ = [
    "BlochLaw",
    "FMR_Power",
    "Inverse_Kittel",
    "KittelEquation",
    "Langevin",
    "blochLaw",
    "fmr_power",
    "inverse_kittel",
    "kittelEquation",
    "langevin",
]

import numpy as np
import scipy.constants as cnst
from scipy.special import zeta, gamma
from scipy.constants import k, mu_0, e, electron_mass, hbar
from scipy.signal import savgol_filter

from lmfit import Model
from lmfit.models import update_param_vals

mu_B = cnst.physical_constants["Bohr magneton"][0]


def blochLaw(T, Ms, Tc):
    r"""Bloch's law for spontaneous magnetism at low temperatures.

    Args:
        T (array like):
            Temperatures at which M has been measured (Probably in K)
        Ms (float):
            The magnetisation at 0K (in whtaever units M is being measured in)
        Tc (float):
            Curie temperature in K

    Returns:
        (array like):
            Calculated values of M

    Notes:
        This fits the Bloch Law as given by:

            :math:`M(T) = M_s\left[1 - \left(\frac{T}{T_C}\right)^\frac{3}{2}\right]`

        This expression is only really valid in the low temperature regime.
    """
    return Ms * (1 - (T / Tc) ** 1.5)


def langevin(H, M_s, m, T):
    r"""Langevin function for paramagnetic M-H loops.

    Args:
        H (array): The applied magnetic field
        M_s (float): Saturation magnetisation
        m (float) is the moment of a cluster
        T (float): Temperature

    Returns:
        Magnetic Momemnts (array).

    Note:
        The Langevin Function is :math:`\coth(\frac{\mu_0HM_s}{k_BT})-\frac{k_BT}{\mu_0HM_s}`.

    Example:
        .. plot:: samples/Fitting/langevin.py
            :include-source:
            :outname: langevin
    """
    x = mu_0 * H * m / (k * T)
    n = M_s / m

    return m * n * (1.0 / np.tanh(x) - 1.0 / x)


def kittelEquation(H, g, M_s, H_k):
    r"""Kittel Equation for finding ferromagnetic resonance peak in frequency with field.

    Args:
        H (array): Magnetic fields in A/m
        g (float): g factor for the gyromagnetic radius
        M_s (float): Magnetisation of sample in A/m
        H_k (float): Anisotropy field term (including demagnetising factors) in A/m

    Returns:
        Reesonance peak frequencies in Hz

    :math:`f = \frac{\gamma \mu_{0}}{2 \pi} \sqrt{\left(H + H_{k}\right) \left(H + H_{k} + M_{s}\right)}`

    Example:
        .. plot:: samples/Fitting/kittel.py
            :include-source:
            :outname: kittel
    """
    gamma = g * cnst.e / (2 * cnst.m_e)
    return (cnst.mu_0 * gamma / (2 * np.pi)) * np.sqrt((H + H_k) * (H + H_k + M_s))


def inverse_kittel(f, g, M_s, H_k):
    r"""Rewritten Kittel equation for finding ferromagnetic resonsance in field with frequency.

    Args:
        f (array): Resonance Frequency in Hz
        g (float): g factor for the gyromagnetic radius
        M_s (float): Magnetisation of sample in A/m
        H_k (float): Anisotropy field term (including demagnetising factors) in A/m

    Returns:
        Reesonance peak frequencies in Hz

    Notes:
        In practice one often measures FMR by sweepign field for constant frequency and then locates the
        peak in H by fitting a suitable Lorentzian type peak. In this case, one returns a
        :math:`H_{res}\pm \Delta H_{res}`
        In order to make use of this data with :py:meth:`Stoner.Analysis.AnalysisMixin.lmfit` or
        :py:meth:`Stoner.Analysis.AnalysisMixin.curve_fit`
        it makes more sense to fit the Kittel Equation written in terms of H than frequency.

       :math:`H_{res}=- H_{k} - \frac{M_{s}}{2} + \frac{1}{2 \gamma \mu_{0}} \sqrt{M_{s}^{2}
       \gamma^{2} \mu_{0}^{2} + 16 \pi^{2} f^{2}}`
    """
    gamma = g * cnst.e / (2 * cnst.m_e)
    return -H_k - M_s / 2 + np.sqrt(M_s**2 * gamma**2 * cnst.mu_0**2 + 16 * np.pi**2 * f**2) / (2 * gamma * cnst.mu_0)


def fmr_power(H, H_res, Delta_H, K_1, K_2):
    r"""Combine a Lorentzian and differential Lorenztion peak as measured in an FMR experiment.

    Args:
        H (array) magnetic field data
        H_res (float): Resonance field of peak
        Delta_H (float): Preak wideth
        K_1, K_2 (float): Relative weighting of each component.

    Returns:
        Array of model absorption values.

    :math:`\frac{4 \Delta_{H} K_{1} \left(H - H_{res}\right)}{\left(\Delta_{H}^{2} + 4 \left(H - H_{res}
    \right)^{2}\right)^{2}} - \frac{K_{2} \left(\Delta_{H}^{2} - 4 \left(H - H_{res}\right)^{2}\right)}{
    \left(\Delta_{H}^{2} + 4 \left(H - H_{res}\right)^{2}\right)^{2}}`
    """
    return (
        4 * Delta_H * K_1 * (H - H_res) / (Delta_H**2 + 4 * (H - H_res) ** 2) ** 2
        - K_2 * (Delta_H**2 - 4 * (H - H_res) ** 2) / (Delta_H**2 + 4 * (H - H_res) ** 2) ** 2
    )


class BlochLaw(Model):
    r"""Bloch's law for spontaneous magnetism at low temperatures.

    Args:
        T (array like):
            Temperature (K)
        g (float):
            Lande g-factor
        A(float):
            The echange stiffness (Jm^{-1})
        Ms(float):
            Saturation moment (Am^{-1})


    Returns:
        (array like):
            Magnetisation values corresponding to the given temperatures.

    Notes:
        Calculates :math:`1 - \frac{\left((\Gamma(3/2) \zeta(3/2)\right)}
                                          {(4\pi^2)}  (v_{ws} / S) (k_B T / D)^{3/2}`

        This is the bulk version of Bloch's law which is not fully correct for thin films. Model adapted from code
        by Dr Joseph Barker <j.barker@leeds.ac.uk>
    """

    display_names = ["g", "A", "M_s"]
    units = ["", "Jm^{-1}", "Am^{-1}"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(self.blochs_law_bulk, *args, **kwargs)
        self.set_param_hint("g", vary=False, value=2.0)
        self.set_param_hint("A", vary=True, min=0)
        self.set_param_hint("Ms", vary=True, min=0)
        self.prefactor = gamma(1.5) * zeta(1.5) / (4 * np.pi**2)

    def blochs_law_bulk(self, T, g, A, Ms):
        r"""Bloch's Law in bulk systems.

        Args:
            T (array like):
                Temperature (K)
            g (float):
                Lande g-factor
            A(float):
                The echange stiffness (Jm^{-1})
            Ms(float):
                Saturation moment (Am^{-1})

        Returns:
            (array like):
                Magnetisation values corresponding to the given temperatures.

        Notes:
            Calculates :math:`1 - \frac{\left((\Gamma(3/2) \zeta(3/2)\right)}
                                              {(4\pi^2)}  (v_{ws} / S) (k_B T / D)^{3/2}`
        """
        Tp = (k * Ms * T / (2 * g * mu_B * A)) ** 1.5
        prefactor = self.prefactor * g * mu_B
        ret = Ms - (prefactor * Tp)
        return ret

    def guess(self, data, x=None, **kwargs):
        """Guess some starting values."""
        Ms = data.max() * 1.001
        if x is not None:
            Mx_max = data[x.argmax()]
            y = ((Ms - Mx_max) / (self.prefactor * 2 * mu_B)) ** (2.0 / 3)
            Ag = Ms * k * x.max() / (y * 2 * mu_B)
        else:
            Ag = 1.0

        guesses = {}
        guesses["Ms"] = Ms
        if self.param_hints["A"]["vary"]:
            A = Ag / self.param_hints["g"]["value"]
            guesses["A"] = A
        elif self.param_hints["g"]["vary"]:
            g = Ag / self.param_hints["A"]["value"]
            guesses["g"] = g

        pars = self.make_params(**guesses)
        return update_param_vals(pars, self.prefix, **kwargs)


class BlochLawThin(Model):
    r"""Bloch's law for spontaneous magnetism at low temperatures - thin film version.

    Args:
        T (array like):
            Temperature (K)
        g (float):
            Lande g-factor
        A(float):
            The echange stiffness (Jm^{-1})
        Ms(float):
            Saturation moment (Am^{-1})


    Returns:
        (array like):
            Magnetisation values corresponding to the given temperatures.

    Notes:
        Calculates :math:`1 - \frac{\left((\Gamma(3/2) \zeta(3/2)\right)}
                                          {(4\pi^2)}  (v_{ws} / S) (k_B T / D)^{3/2}`

        This is the bulk version of Bloch's law which is not fully correct for thin films. Model adapted from code
        by Dr Joseph Barker <j.barker@leeds.ac.uk>
    """

    def __init__(self, *args, **kwargs):
        """Initialise the model."""
        super().__init__(self.blochs_law_thinfilm, *args, **kwargs)
        self.set_param_hint("g", vary=False, value=2.0)
        self.set_param_hint("A", vary=True, min=0)
        self.set_param_hint("Ms", vary=True, min=0)
        self.prefactor = gamma(1.5) * zeta(1.5) / (4 * np.pi**2)

    def blochs_law_thinfilm(self, T, D, Bz, S, v_ws, a, nz):
        r"""Thin film version of Blopch's Law.

        Parameters:
            T (array):
                Temperature (K)
            D (float):
                Spin wave stiffness
            Bz (float):
                Applied magnetic field data measured in.
            S (float):
                Spin number
            v_ws (float):
                Wigner-Seitz volume (volume per atom)
            a (float):
                Lattice parameter
            nz (int):
                Number of atomic layers in the thin film.

        Returns:
            (array):
                normalised M_s values.

        Notes:
            Bloch's Law is derived from integrating the magnon spectrum over the k-space of the sample. In a thin film
            one can only integrate over the x-y plane, in the z direction one needs to do an explicit sum over the
            atomic layers in the z direction.

            This model was derived by Dr Joseph Barker <j.barker@leeds.ac.uk>

            This is solving:

                :math:`1 + \frac{v_{ws}}{S}\frac{a}{4 \pi n_z}\frac{k_B T}{D}\sum^{n_z-1}_{m=0}
                        ln\left[1-\exp\left( -\frac{1}{k_B T}\left(g\mu_B B_z+D\left(\frac{m \pi}{a(n_z-1)}\right)^2
                                                                   \right)\right) \right]`
        """
        kz_sum = sum(
            np.log(1.0 - np.exp(-(D * (m * np.pi / ((nz - 1) * a)) ** 2 + Bz) / (k * T))) for m in range(0, nz - 1)
        )
        return 1.0 + (v_ws / (4 * np.pi * S * (nz - 1) * a)) * (k * T / D) * kz_sum


class Langevin(Model):
    r"""The Langevin function for paramagnetic M-H loops.

    Args:
        H (array): The applied magnetic field
        M_s (float): Saturation magnetisation
        m (float): is the moment of a single cluster
        T (float): Temperature

    Returns:
        Magnetic Momemnts (array).

    Note:
        The Langevin Function is :math:`\coth(\frac{\mu_0HM_s}{k_BT})-\frac{k_BT}{\mu_0HM_s}`.

    Example:
        .. plot:: samples/Fitting/langevin.py
            :include-source:
            :outname: langevin-class
    """

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(langevin, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        r"""Guess some starting values.

        M_s is taken as half the difference of the range of thew M data,
        we can find m/T from the susceptibility :math:`chi= M_s \mu_o m / kT`
        """
        M_s = (np.max(data) - np.min(data)) / 2.0
        if x is not None:
            d = np.sort(np.row_stack((x, data)))
            dd = savgol_filter(d, 7, 1)
            yd = dd[1] / dd[0]
            chi = np.interp(np.array([0]), d[0], yd)[0]
            mT = chi / M_s * (k / mu_0)
            # Assume T=150K for no good reason
            m = mT * 150
        else:
            m = 1e6 * (e * hbar) / (2 * electron_mass)  # guess 1 million Bohr Magnetrons
        T = 150
        pars = self.make_params(M_s=M_s, m=m, T=T)
        return update_param_vals(pars, self.prefix, **kwargs)


class KittelEquation(Model):
    r"""Kittel Equation for finding ferromagnetic resonance peak in frequency with field.

    Args:
        H (array): Magnetic fields in A/m
        g (float): h g factor for the gyromagnetic radius
        M_s (float): Magnetisation of sample in A/m
        H_k (float): Anisotropy field term (including demagnetising factors) in A/m

    Returns:
        Reesonance peak frequencies in Hz

    :math:`f = \frac{\gamma \mu_{0}}{2 \pi} \sqrt{\left(H + H_{k}\right) \left(H + H_{k} + M_{s}\right)}`

    Example:
        .. plot:: samples/Fitting/kittel.py
            :include-source:
            :outname: kittel-class
    """

    display_names = ["g", "M_s", "H_k"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(kittelEquation, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess parameters as gamma=2, H_k=0, M_s~(pi.f)^2/(mu_0^2.H)-H."""
        g = 2
        H_k = 100
        gamma = g * cnst.e / (2 * cnst.m_e)
        M_s = (4 * np.pi**2 * data**2 - gamma**2 * cnst.mu_0**2 * (x**2 + 2 * x * H_k + H_k**2)) / (
            gamma**2 * cnst.mu_0**2 * (x + H_k)
        )
        M_s = np.mean(M_s)

        pars = self.make_params(g=g, M_s=M_s, H_k=H_k)
        pars["M_s"].min = 0
        pars["g"].min = g / 100
        pars["H_k"].min = 0
        pars["H_k"].max = M_s.max()
        return update_param_vals(pars, self.prefix, **kwargs)


class Inverse_Kittel(Model):
    r"""Kittel Equation for finding ferromagnetic resonance peak in frequency with field.

    Args:
        H (array): Magnetic fields in A/m
        g (float): h g factor for the gyromagnetic radius
        M_s (float): Magnetisation of sample in A/m
        H_k (float): Anisotropy field term (including demagnetising factors) in A/m

    Returns:
        Reesonance peak frequencies in Hz

    :math:`f = \frac{\gamma \mu_{0}}{2 \pi} \sqrt{\left(H + H_{k}\right) \left(H + H_{k} + M_{s}\right)}`

    """

    display_names = ["g", "M_s", "H_k"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(inverse_kittel, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Guess parameters as gamma=2, H_k=0, M_s~(pi.f)^2/(mu_0^2.H)-H."""
        g = 2
        H_k = 100
        gamma = g * cnst.e / (2 * cnst.m_e)
        M_s = (4 * np.pi**2 * x**2 - gamma**2 * cnst.mu_0**2 * (data**2 + 2 * data * H_k + H_k**2)) / (
            gamma**2 * cnst.mu_0**2 * (data + H_k)
        )
        M_s = np.mean(M_s)

        pars = self.make_params(g=g, M_s=M_s, H_k=H_k)
        pars["M_s"].min = 0
        pars["g"].min = g / 100
        pars["H_k"].min = 0
        pars["H_k"].max = M_s.max()
        return update_param_vals(pars, self.prefix, **kwargs)


class FMR_Power(Model):
    r"""Combine a Lorentzian and differential Lorenztion peak as measured in an FMR experiment.

    Args:
        H (array) magnetic field data
        H_res (float): Resonance field of peak
        Delta_H (float): Preak wideth
        K_1, K_2 (float): Relative weighting of each component.

    Returns:
        Array of model absorption values.

    :math:`\frac{4 \Delta_{H} K_{1} \left(H - H_{res}\right)}{\left(\Delta_{H}^{2} + 4 \left(H - H_{res}
    \right)^{2}\right)^{2}} - \frac{K_{2} \left(\Delta_{H}^{2} - 4 \left(H - H_{res}\right)^{2}\right)}{\left(
    \Delta_{H}^{2} + 4 \left(H - H_{res}\right)^{2}\right)^{2}}`
    """

    display_names = ["H_{res}", r"\Delta_H", "K_1", "K_2"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(fmr_power, *args, **kwargs)

    def guess(self, data, x=None, **kwargs):
        r"""Guess parameters as :math:`\gamma=2, H_k=0, M_s~(\pi f)^2/\mu_0^2.H)-H`."""
        if x is None:
            x = np.linspace(1, len(data), len(data) + 1)

        x1 = x[np.argmax(data)]
        x2 = x[np.argmin(data)]
        Delta_H = abs(x1 - x2)
        H_res = (x1 + x2) / 2.0
        y1 = np.max(data)
        y2 = np.min(data)
        dy = y1 - y2
        K_2 = dy * (4 * np.pi * Delta_H**2) / (3 * np.sqrt(3))
        ay = (y1 + y2) / 2
        K_1 = ay * np.pi / Delta_H

        pars = self.make_params(Delta_H=Delta_H, H_res=H_res, K_1=K_1, K_2=K_2)
        pars["K_1"].min = 0
        pars["K_2"].min = 0
        pars["Delta_H"].min = 0
        pars["H_res"].min = np.min(x)
        pars["H_res"].max = np.max(x)
        return update_param_vals(pars, self.prefix, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Fit IV data to various RSJ models."""
# pylint: disable=invalid-name
import os

from Stoner import Data, __home__
from Stoner.analysis.fitting.models.superconductivity import (
    RSJ_Noiseless,
    RSJ_Simple,
)

os.chdir(os.path.join(__home__, "..", "doc", "samples", "Fitting"))

data = Data("data/IV.txt", setas={"x": "Current", "y": "Voltage"})

# Fit data with both versions of the RSJ model
data.lmfit(RSJ_Simple, result=True, header="Simple", prefix="simple")
data.lmfit(RSJ_Noiseless, result=True, header="Noiseless", prefix="noiseless")

# Set column assignments and plot the data and fits
data.setas = {"x": "Current", "y": ["Voltage", "Simple", "Noiseless"]}
data.plot(fmt=["r+", "b-", "g-"])

# Annotate fits
data.annotate_fit(
    RSJ_Simple, prefix="simple", mode="eng", x=0.15, y=0.1, fontsize="small"
)
data.annotate_fit(
    RSJ_Noiseless,
    prefix="noiseless",
    mode="eng",
    x=0.55,
    y=0.1,
    fontsize="small",
)



# -*- coding: utf-8 -*-
"""General file related tools."""
from importlib import import_module
import io
import os
import pathlib
import urllib
from traceback import format_exc
from typing import Union, Sequence, Dict, Type, Tuple, Optional

from ..compat import string_types, path_types, bytes2str, str2bytes
from .widgets import fileDialog
from .classes import subclasses
from ..core.exceptions import StonerLoadError, StonerUnrecognisedFormat
from ..core.base import regexpDict, metadataObject

from ..core.Typing import Filename

__all__ = [
    "file_dialog",
    "get_file_name_type",
    "auto_load_classes",
    "get_mime_type",
    "FileManager",
    "SizedFileManager",
    "URL_SCHEMES",
]

try:
    from magic import Magic as filemagic, MAGIC_MIME_TYPE
except ImportError:
    filemagic = None

URL_SCHEMES = ["http", "https"]


def file_dialog(
    mode: str, filename: Filename, filetype: Union[Type[metadataObject], str], baseclass: Type[metadataObject]
) -> Union[pathlib.Path, Sequence[pathlib.Path], None]:
    """Create a file dialog box for loading or saving ~b DataFile objects.

    Args:
        mode (string):
            The mode of the file operation  'r' or 'w'
        filename (str, Path, bool):
            The starting filename
        filetype (submclass of metadataObject, string):
            The filetype to open with - used to selectr file patterns
        basclass (subclass of metadataObject):
            Type of object we're looking to create'

    Returns:
        (pathlib.PurePath or None):
            A filename to be used for the file operation.
    """
    # Wildcard pattern to be used in file dialogs.

    descs: Dict = {"*.*": "All Files"}
    for p in filetype.patterns:  # pylint: disable=not-an-iterable
        descs[p] = filetype.__name__ + " file"
    for c in subclasses(baseclass):  # pylint: disable=E1136, E1133
        for p in subclasses(baseclass)[c].patterns:  # pylint: disable=unsubscriptable-object
            if p in descs:
                descs[p] += (
                    ", " + subclasses(baseclass)[c].__name__ + " file"  # pylint: disable=E1136
                )  # pylint: disable=unsubscriptable-object
            else:
                descs[p] = subclasses(baseclass)[c].__name__ + " file"  # pylint: disable=unsubscriptable-object

    patterns = descs

    if isinstance(filename, string_types):
        filename = pathlib.Path(filename)

    if filename is not None and not isinstance(filename, bool):
        dirname = filename.parent
        filename = filename.name
    else:
        filename = ""
        dirname = ""
    if "r" in mode:
        mode = "OpenFile"
    elif "w" in mode:
        mode = "SaveFile"
    else:
        mode = "SelectDirectory"
    filename = fileDialog.openDialog(start=dirname, mode=mode, patterns=patterns)
    return filename if filename else None


def get_file_name_type(
    filename: Filename, filetype: Union[Type[metadataObject], str], parent: Type[metadataObject]
) -> Tuple[pathlib.PurePath, Type[metadataObject]]:
    """Rationalise a filename and filetype."""
    if isinstance(filename, string_types):
        filename = pathlib.Path(filename)
    if isinstance(filetype, string_types):  # We can specify filetype as part of name
        try:
            filetype = regexpDict(subclasses(parent))[filetype]  # pylint: disable=E1136
        except KeyError:
            parts = filetype.split(".")
            mod = ".".join(parts[:-1])
            try:
                mod = import_module(mod)
                filetype = getattr(mod, parts[-1])
            except (ImportError, AttributeError) as err:
                raise ValueError(f"Unable to import {filetype}") from err
            if not issubclass(filetype, parent):
                raise ValueError(f"{filetype} is  not a subclass of DataFile.")
    if filename is None or (isinstance(filename, bool) and not filename):
        if filetype is None:
            filetype = parent
        filename = file_dialog("r", filename, filetype, parent)
    elif isinstance(filename, io.IOBase):  # Opened file
        filename = filename.name
    try:
        if not filename.exists():
            raise IOError(f"Cannot find {filename} to load")
    except AttributeError as err:
        raise IOError(f"Unable to tell if file exists - {type(filename)}") from err
    return filename, filetype


def auto_load_classes(
    filename: Filename,
    baseclass: Type[metadataObject],
    debug: bool = False,
    args: Optional[Tuple] = None,
    kargs: Optional[Dict] = None,
) -> Type[metadataObject]:
    """Work through subclasses of parent to find one that will load this file."""
    mimetype = get_mime_type(filename, debug=debug)
    args = args if args is not None else ()
    kargs = kargs if kargs is not None else {}
    for cls in subclasses(baseclass).values():  # pylint: disable=E1136, E1101
        cls_name = cls.__name__
        if debug:
            print(cls_name)
        try:
            if mimetype is not None and mimetype not in cls.mime_type:  # short circuit for non-=matching mime-types
                if debug:
                    print(f"Skipping {cls_name} due to mismatcb mime type {cls.mime_type}")
                continue
            test = cls()
            if "_load" not in cls.__dict__:  # No local _load method
                continue
            if debug and filemagic is not None:
                print(f"Trying: {cls_name} =mimetype {test.mime_type}")

            test = test._load(filename, auto_load=False, *args, **kargs)
            if test is None:
                raise SyntaxError(f"Class {cls_name}'s _load returned None !!")
            try:
                kargs = test._kargs
                delattr(test, "_kargs")
            except AttributeError:
                pass

            if debug:
                print("Passed Load")
            if isinstance(test, metadataObject):
                test["Loaded as"] = cls_name
            if debug:
                print(f"Test metadata: {test.metadata}")

            break
        except StonerLoadError as e:
            if debug:
                print(f"Failed Load: {e}")
            continue
        except UnicodeDecodeError:
            print(f"{cls, filename} Failed with a uncicode decode error for { format_exc()}\n")
            continue
    else:
        raise StonerUnrecognisedFormat(
            f"Ran out of subclasses to try and load {filename} (mimetype={mimetype}) as."
            + f" Recognised filetype are:{list(subclasses(baseclass).keys())}"  # pylint: disable=E1101
        )
    return test


def get_mime_type(filename: Union[pathlib.Path, str], debug: bool = False) -> Optional[str]:
    """Get the mime type of the file if filemagic is available."""
    if (
        filemagic is not None
        and isinstance(filename, path_types)
        and urllib.parse.urlparse(str(filename)).scheme not in URL_SCHEMES
    ):
        with filemagic(flags=MAGIC_MIME_TYPE) as m:
            mimetype = m.id_filename(str(filename))
        if debug:
            print(f"Mimetype:{mimetype}")
    else:
        mimetype = None
    return mimetype


class FileManager:
    """Simple context manager that allows opening files or working with already open string buffers."""

    def __init__(self, filename, *args, **kargs):
        """Store the parameters passed to the context manager."""
        self.filename = filename
        self.args = args
        self.kargs = kargs
        self.file = None
        self.binary = len(args) > 0 and args[0].endswith("b")
        if isinstance(filename, path_types):
            parsed = urllib.parse.urlparse(str(filename))
            if parsed.scheme not in URL_SCHEMES:
                filename = pathlib.Path(filename)
            else:
                filename = urllib.request.urlopen(filename)
        if isinstance(filename, path_types):
            self.mode = "open"
        elif isinstance(filename, io.IOBase):
            if not hasattr(filename, "response"):
                if self.binary:
                    self.mode = "bytes"
                    self.filename = str2bytes(filename.read())
                else:
                    self.filename = bytes2str(filename.read())
                    self.mode = "text"
                filename.response = self.filename
            else:
                if self.binary:
                    self.mode = "bytes"
                    self.filename = str2bytes(filename.response)
                else:
                    self.filename = bytes2str(filename.response)
                    self.mode = "text"
        elif isinstance(filename, bytes):
            if (len(args) > 0 and args[0][-1] == "b") or self.kargs.pop("mode", "").endswith("b"):
                self.filename = filename
                self.mode = "bytes"
            else:
                self.filename = bytes2str(filename)
                self.mode = "text"
        else:
            raise TypeError(f"Unrecognised filename type {type(filename)}")

    def __enter__(self):
        """Either open the file or reset the buffer."""
        if self.mode == "open":
            if len(self.args) > 0 and "b" not in self.args[0]:
                self.kargs.setdefault("encoding", "utf-8")
            self.file = open(self.filename, *self.args, **self.kargs)
        elif self.mode == "text":
            self.file = io.StringIO(self.filename)
        elif self.mode == "bytes":
            self.file = io.BytesIO(self.filename)
        elif self.mode in ["bytesio", "textio"]:
            self.file = self.filename
        else:
            raise TypeError(f"Unrecognised filename type {type(self.filename)}")
        return self.file

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Close the open file, or reset the buffer position."""
        if not self.file.closed and self.file.seekable():
            self.file.seek(0)
        if self.mode == "open":
            self.file.close()


class SizedFileManager(FileManager):
    """Context manager that figures out the size of the file as well as opening it."""

    def __enter__(self):
        """Add the file length information to the context variable."""
        super().__enter__()
        if self.mode == "open":
            length = os.stat(self.filename).st_size
        elif self.mode in ["textio", "bytesio"]:
            if self.file.seekable():
                pos = self.file.tell()
                self.file.seek(0, 2)
                length = self.file.tell()
                self.file.seek(pos)
            else:
                length = -1
        elif self.mode in ["text", "bytes"]:
            length = len(self.filename)
        else:
            length = len(self.file)
        return self.file, length



"""test_Core.py
Created on Tue Jan 07 22:05:55 2014

@author: phygbu
"""

import sys
import os.path as path
import numpy as np
import re
import pandas as pd
import pytest
from numpy import all, sqrt, nan
from collections.abc import MutableMapping

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
old_path = []
from Stoner import Data, __home__
from Stoner.Core import metadataObject
import Stoner.compat


def mask_func(r):
    return np.abs(r.q) < 0.25 * np.pi


datadir = path.join(pth, "sample-data")


def setup_module():
    global old_path
    old_path = sys.path
    sys.path.insert(0, pth)


def teardown_module():
    global old_path
    sys.path = old_path


def setup_function(function):
    global selfd, selfd1, selfd2, selfd3, selfd4

    selfd = Data(path.join(path.dirname(__file__), "CoreTest.dat"), setas="xy")
    selfd2 = Data(path.join(__home__, "..", "sample-data", "TDI_Format_RT.txt"))
    selfd3 = Data(path.join(__home__, "..", "sample-data", "New-XRay-Data.dql"))
    selfd4 = Data(path.join(__home__, "..", "sample-data", "Cu_resistivity_vs_T.txt"))


def test_repr():
    global selfd, selfd1, selfd2, selfd3, selfd4
    header = """==============  ======  ======
TDI Format 1.5  X-Data  Y-Data
    index       0 (x)   1 (y)
==============  ======  ======"""
    repr_header = """=============================  ========  ========
TDI Format 1.5                   X-Data    Y-Data
index                             0 (x)     1 (y)
=============================  ========  ========"""
    assert selfd.header, header == "Header output changed."
    assert "\n".join(repr(selfd).split("\n")[:4]) == repr_header, "Representation gave unexpected header."
    assert len(repr(selfd2).split("\n")) == 261, "Long file representation failed"
    assert len(selfd2._repr_html_().split("\n")) == 263, "Long file representation failed"


def test_base_class():
    global selfd, selfd1, selfd2, selfd3, selfd4
    m = metadataObject()
    m.update(selfd2.metadata)
    assert isinstance(m, MutableMapping), "metadataObject is not a MutableMapping"
    assert len(m) == len(selfd2.metadata), "metadataObject length failure."
    assert len(m[".*"]) == len(m), "Failure to index by regexp."
    for k in selfd2.metadata:
        assert m[k] == selfd2.metadata[k], f"Failure to have equal keys for {k}"
        m[k] = k
        assert m[k] == k, f"Failure to set keys for {k}"
        del m[k]
    assert len(m) == 0, "Failed to delete from metadataObject."


def test_constructor():
    """Constructor Tests"""
    global selfd, selfd1, selfd2, selfd3, selfd4
    d = Data()
    assert d.shape == (1, 0), "Bare constructor failed"
    d = Data(selfd)
    assert np.all(d.data == selfd.data), "Constructor from DataFile failed"
    d = Data([np.ones(100), np.zeros(100)])
    assert d.shape == (100, 2), "Constructor from iterable list of nd array failed"
    d = Data([np.ones(100), np.zeros(100)], ["X", "Y"])
    assert d.column_headers == ["X", "Y"], f"Failed to set column headers in constructor: {d.column_headers}"
    c = np.zeros(100)
    d = Data({"X-Data": c, "Y-Data": c, "Z-Data": c})
    assert d.shape == (100, 3), "Construction from dictionary of columns failed."
    d = selfd.clone
    df = d.to_pandas()
    e = Data(df)
    assert d == e, "Roundtripping through Pandas DataFrame failed."
    d = selfd.clone
    e = Data(d.dict_records)
    e.metadata = d.metadata
    assert d == e, "Copy from dict records failed."
    d = Data(pd.DataFrame(np.zeros((100, 3))))
    assert d.column_headers == ["Column 0:0", "Column 1:1", "Column 2:2"]
    d = Data(np.ones(100), np.zeros(100), np.zeros(100), selfd.filename)
    assert d == selfd, "Multiple argument constructor dind't find it's filename"
    with pytest.raises(TypeError):
        e = d(setas=34)
    with pytest.raises(TypeError):
        e = d(column_headers=34)

    with pytest.raises(AttributeError):
        d = Data(bad_attr=False)
    with pytest.raises(TypeError):
        d = Data(column_headers=43)
    with pytest.raises(TypeError):
        d = Data(object)


def test_column():
    global selfd, selfd1, selfd2, selfd3, selfd4
    for i, c in enumerate(selfd.column_headers):
        # Column function checks
        assert all(selfd.data[:, i] == selfd.column(i)), f"Failed to Access column {i} by index"
        assert all(selfd.data[:, i] == selfd.column(c)), f"Failed to Access column {i} by name"
        assert all(selfd.data[:, i] == selfd.column(c[0:3])), f"Failed to Access column {i} by partial name"

    # Check that access by list of strings returns multpiple columns
    assert all(
        selfd.column(selfd.column_headers) == selfd.data
    ), "Failed to access all columns by list of string indices"
    assert all(
        selfd.column([0, selfd.column_headers[1]]) == selfd.data
    ), "Failed to access all columns by mixed list of indices"

    # Check regular expression column access
    assert all(
        selfd.column(re.compile(r"^X\-"))[:, 0] == selfd.column(0)
    ), "Failed to access column by regular expression"
    # Check attribute column access
    assert all(selfd.X == selfd.column(0)), "Failed to access column by attribute name"


def test_deltions():
    global selfd, selfd1, selfd2, selfd3, selfd4
    ch = ["{}-Data".format(chr(x)) for x in range(65, 91)]
    data = np.zeros((100, 26))
    metadata = dict([("Key 1", True), ("Key 2", 12), ("Key 3", "Hello world")])
    selfdd = Data(metadata)
    selfdd.data = data
    selfdd.column_headers = ch
    selfdd.setas = "3.x3.y3.z"
    selfrepr_string = (
        "==========================  ========  =======  ========  =======  ========  ========\n"
        + "TDI Format 1.5                D-Data   ....      H-Data   ....      Y-Data    Z-Data\n"
        + "index                          3 (x)              7 (y)                 24        25\n"
        + "==========================  ========  =======  ========  =======  ========  ========\n"
        + "Key 1{Boolean}= True               0                  0                  0         0\n"
        + "Key 2{I32}= 12                     0  ...             0  ...             0         0\n"
        + "Key 3{String}= Hello world         0  ...             0  ...             0         0\n"
        + "Stoner.class{String}= Data         0  ...             0  ...             0         0\n"
        + "...                                0  ...             0  ...             0         0"
    )
    assert (
        "\n".join(repr(selfdd).split("\n")[:9]) == selfrepr_string
    ), "Representation with interesting columns failed."
    del selfdd["Key 1"]
    assert len(selfdd.metadata) == 3, "Deletion of metadata failed."
    del selfdd[20:30]
    assert selfdd.shape == (90, 26), "Deleting rows directly failed."
    selfdd.del_column("Q")
    assert selfdd.shape == (90, 25), "Deleting rows directly failed."
    selfdd %= 3
    assert selfdd.shape == (90, 24), "Deleting rows directly failed."
    selfdd.setas = "x..y..z"
    selfdd.del_column(selfdd.setas.not_set)
    assert selfdd.shape == (90, 3), "Deleting rows directly failed."
    selfdd.mask[50, 1] = True
    selfdd.del_column()
    assert selfdd.shape == (90, 2), "Deletion of masked rows failed."
    selfd5 = Data(np.ones((10, 10)))
    selfd5.column_headers = ["A"] * 5 + ["B"] * 5
    selfd5.del_column("A", duplicates=True)
    assert selfd5.shape == (10, 6), "Failed to delete columns with duplicates True and col specified."
    selfd5 = Data(np.ones((10, 10)))
    selfd5.column_headers = list("ABCDEFGHIJ")
    selfd5.setas = "..x..y"
    selfd5.del_column(True)
    assert selfd5.column_headers == ["C", "F"], "Failed to delete columns with col=True"


def test_attributes():
    """Test various attribute accesses,"""
    global selfd, selfd1, selfd2, selfd3, selfd4
    header = (
        "==============  ======  ======\n"
        + "TDI Format 1.5  X-Data  Y-Data\n"
        + "    index       0 (x)   1 (y)\n"
        + "==============  ======  ======"
    )
    assert selfd.shape == (100, 2), "shape attribute not correct."
    assert selfd.dims == 2, "dims attribute not correct."
    assert selfd.header == header, "Shorter header not as expected"
    assert all(selfd.T == selfd.data.T), "Transpose attribute not right"
    assert all(selfd.x == selfd.column(0)), "x attribute quick access not right."
    assert all(selfd.y == selfd.column(1)), "y attribute not right."
    assert all(selfd.q == np.arctan2(selfd.data[:, 0], selfd.data[:, 1])), "Calculated theta attribute not right."
    assert all(sqrt(selfd[:, 0] ** 2 + selfd[:, 1] ** 2) == selfd.r), "Calculated r attribute not right."
    assert selfd2.records._[5]["Column 2"] == selfd2.data[5, 2], "Records and as array attributes problem"
    d = Data(np.ones((10, 2)), ["A", "A"])
    assert d.records.dtype == np.dtype(
        [("A_1", "<f8"), ("A", "<f8")]
    ), ".records with identical column headers handling failed"
    e = selfd.clone
    e.data = np.array([1, 1, 1])
    e.T = selfd.T
    e.column_headers = selfd.column_headers
    assert e == selfd, "Writingto transposed data failed"
    assert selfd.xcol == 0
    assert selfd.ycol == [1]
    with pytest.raises(AttributeError):
        selfd.zcol is None
    selfd.setas = "xz"
    assert selfd.zcol == [1]
    selfd.x = np.ones((100, 1))
    assert all(selfd.x == np.ones(100))
    selfd.x = np.ones((1, 100))
    assert all(selfd.x == np.ones(100))
    with pytest.raises(RuntimeError):
        selfd.x = np.ones((2, 2))


def test_mask():
    global selfd, selfd1, selfd2, selfd3, selfd4
    selflittle = Data()
    p = np.linspace(0, np.pi, 91)
    q = np.linspace(0, 2 * np.pi, 91)
    r = np.cos(p)
    x = r * np.sin(q)
    y = r * np.cos(q)
    selflittle.data = np.column_stack((x, y, r))
    selflittle.setas = "xyz"
    selflittle.mask = mask_func
    selflittle.mask[:, 2] = True
    assert len(repr(selflittle).split("\n")) == len(selflittle) + 5, "Non shorterned representation failed."
    selflittle.mask = lambda r: np.abs(r) > 0.5
    selflittle.del_rows()
    assert selflittle.shape == (30, 3)


def test_iterators():
    global selfd, selfd1, selfd2, selfd3, selfd4
    for i, c in enumerate(selfd.columns()):
        assert all(selfd.column(i) == c), "Iterating over DataFile.columns not the same as direct indexing column"
    for j, r in enumerate(selfd.rows()):
        assert all(selfd[j] == r), "Iteratinf over DataFile.rows not the same as indexed access"
    for k, r in enumerate(selfd):
        pass
    assert j == k, "Iterating over DataFile not the same as DataFile.rows"
    assert selfd.data.shape == (j + 1, i + 1), "Iterating over rows and columns not the same as data.shape"


def test_dir():
    global selfd, selfd1, selfd2, selfd3, selfd4
    assert selfd.dir("S") == ["Stoner.class"], f"Dir method failed: dir was {selfd.dir()}"
    bad_keys = set(
        [
            "__class_getitem__",
            "__metaclass__",
            "iteritems",
            "iterkeys",
            "itervalues",
            "__ge__",
            "__gt__",
            "__init_subclass__",
            "__le__",
            "__lt__",
            "__reversed__",
            "__slots__",
            "_abc_negative_cache",
            "_abc_registry",
            "_abc_negative_cache_version",
            "_abc_cache",
            "_abc_impl",
            "__annotations__",
            "__getstate__",
        ]
    )
    attrs = set(dir(selfd)) - bad_keys
    assert len(attrs) == 247, "DataFile.__dir__ failed."
    selfd.setas.clear()
    attrs = set(dir(selfd)) - bad_keys
    assert len(attrs) == 245, "DataFile.__dir__ failed."


def test_filter():
    global selfd, selfd1, selfd2, selfd3, selfd4
    selfd._push_mask()
    ix = np.argmax(selfd.x)
    selfd.filter(lambda r: r.x <= 50)
    assert np.max(selfd.x) == 50, "Failure of filter method to set mask"
    assert np.isnan(selfd.x[ix]), "Failed to mask maximum value"
    selfd._pop_mask()
    assert selfd2.select(Temp__not__gt=150).shape == (839, 3), "Seect method failure."
    assert selfd.select(lambda r: r.x < 30) == selfd.select(X__lt=30), "Select  method as callable failed."
    assert selfd.select(__=lambda r: r.x < 30) == selfd.select(X__lt=30), "Select  method as callable failed."


def test_properties():
    global selfd, selfd1, selfd2, selfd3, selfd4
    selflittle = Data()
    p = np.linspace(0, np.pi, 91)
    q = np.linspace(0, 2 * np.pi, 91)
    r = np.cos(p)
    x = r * np.sin(q)
    y = r * np.cos(q)
    selflittle.data = np.column_stack((x, y, r))
    selflittle.setas = "xyz"
    q_ang = np.round(selflittle.q / np.pi, decimals=2)
    p_ang = np.round(selflittle.p / np.pi, decimals=2)
    assert np.max(q_ang) == 1.0, "Data.q failure"
    assert np.max(p_ang) == 0.5, "Data.p failure"
    assert np.min(p_ang) == -0.5, "Data.p failure"


def test_methods():
    global selfd, selfd1, selfd2, selfd3, selfd4
    d = selfd.clone
    d &= np.where(d.x < 50, 1.0, 0.0)
    d.rename(2, "Z-Data")
    d.setas = "xyz"
    assert all(d.unique(2) == np.array([0, 1])), f"Unique values failed: {d.unique(2)}"
    d = selfd.clone
    d.insert_rows(10, np.zeros((2, 2)))
    assert len(d) == 102, "Failed to inert extra rows"
    assert d[9, 0] == 10 and d[10, 0] == 0 and d[12, 0] == 11, "Failed to insert rows properly."
    d = selfd.clone
    d.add_column(np.ones(len(d)), replace=False, header="added")
    assert d.shape[1] == selfd.shape[1] + 1, "Adding a column with replace=False did not add a column."
    assert np.all(d.data[:, -1] == np.ones(len(d))), "Didn't add the new column to the end of the data."
    assert len(d.column_headers) == len(selfd.column_headers) + 1, "Column headers isn't bigger by one"
    assert d.column_headers == selfd.column_headers + [
        "added",
    ], "Column header not added correctly"
    d = selfd.clone
    d.add_column(selfd.x)
    assert np.all(d.x == d[:, 2]), "Adding a column as a DataArray with column headers didn't work"
    assert (
        d.x.column_headers[0] == d.column_headers[2]
    ), "Adding a column as a DataArray with column headers didn't work"
    e = d.clone
    d.swap_column([(0, 1), (0, 2)])
    assert d.column_headers == [
        e.column_headers[x] for x in [2, 0, 1]
    ], f"Swap column test failed: {d.column_headers}"
    e = selfd(setas="yx")
    assert e.shape == selfd.shape and e.setas[0] == "y", "Failed on a DataFile.__call__ test"
    spl = len(repr(selfd).split("\n"))
    assert spl, 105 == f"Failed to do repr function got {spl} lines"
    e = selfd.clone
    e = e.add_column(e.x, header=e.column_headers[0])
    e.del_column(duplicates=True)
    assert e.shape == (100, 2), "Deleting duplicate columns failed"
    e = selfd2.clone
    e.reorder_columns([2, 0, 1])
    assert e.column_headers == [selfd2.column_headers[x] for x in [2, 0, 1]], "Failed to reorder columns: {}".format(
        e.column_headers
    )
    d = selfd.clone
    d.del_rows(0, 10.0)
    assert d.shape == (99, 2), f"Del Rows with value and column failed - actual shape {d.shape}"
    d = selfd.clone
    d.del_rows(0, (10.0, 20.0))
    assert d.shape == (89, 2), f"Del Rows with tuple and column failed - actual shape {d.shape}"
    d = selfd.clone
    d.mask[::2, 0] = True
    d.del_rows()
    assert d.shape == (50, 2), f"Del Rows with mask set - actual shape {d.shape}"
    d = selfd.clone
    d[::2, 1] = nan
    d.del_nan(1)
    assert d.shape == (50, 2), f"del_nan with explicit column set failed shape was {d.shape}"
    d = selfd.clone
    d[::2, 1] = nan
    d.del_nan(0)
    assert d.shape == (100, 2), f"del_nan with explicit column set and not nans failed shape was {d.shape}"
    d = selfd.clone
    d[::2, 1] = nan
    d.setas = ".y"
    d.del_nan()
    assert d.shape == (50, 2), f"del_nan with columns from setas failed shape was {d.shape}"
    d = selfd.clone
    d2 = selfd.clone
    d2.data = d2.data[::-1, :]
    assert d.sort(reverse=True) == d2, "Sorting reverse not the same as manually reversed data."
    d = selfd.clone
    d.mask[::2] = True
    assert d.count() == 50
    assert d.count(9895) == 1
    assert d.count(100, col="X") == 1
    assert selfd.search("X", [98.0, 100]).shape == (2, 2)
    assert selfd.search("X", 50.1, accuracy=0.2, columns=0) == 50.0
    with pytest.raises(RuntimeError):
        selfd.search("X", "Y")
    assert selfd.section(x=(48, 52), accuracy=0.2).shape == (5, 2)
    assert selfd.section(y=(1000, 2000))[0, 0] == 33.0
    assert selfd.select({"X-Data__lt": 50}).shape == (49, 2)
    assert selfd.select({"X-Data": 50})[0, 0] == 50.0
    dd = selfd.clone
    dd.add_column(dd.x % 2, "Channel")
    fldr = dd.split("Channel")
    assert fldr.shape == (2, {})
    assert dd.split("Channel", lambda d: d.x % 3).shape == (0, {0.0: (3, {}), 1.0: (3, {})})
    dd = selfd.clone
    assert dd.split(lambda d: d.x % 3).shape == (3, {})
    assert len(dd.split(lambda d: 1)) == 1
    shp1 = dd.split(lambda d: np.sqrt(d.x) == np.round(np.sqrt(d.x))).shape
    shp2 = dd.split(lambda d: np.sqrt(d.x) == int(np.sqrt(d.x))).shape
    assert shp1 == shp2


def test_metadata_save():
    global selfd, selfd1, selfd2, selfd3, selfd4
    local = path.dirname(__file__)
    t = np.arange(12).reshape(3, 4)  # set up a test data file with mixed metadata
    t = Data(t)
    t.column_headers = ["1", "2", "3", "4"]
    metitems = [
        True,
        1,
        0.2,
        {"a": 1, "b": "abc"},
        (1, 2),
        np.arange(3),
        [1, 2, 3],
        "abc",  # all types accepted
        r"\\abc\cde",
        1e-20,  # extra tests
        [1, (1, 2), "abc"],  # list with different types
        [[[1]]],  # nested list
        None,  # None value
    ]
    metnames = ["t" + str(i) for i in range(len(metitems))]
    for k, v in zip(metnames, metitems):
        t[k] = v
    t.save(path.join(local, "mixedmetatest.dat"))
    tl = Data(
        path.join(local, "mixedmetatest.txt")
    )  # will change extension to txt if not txt or tdi, is this what we want?
    t2 = selfd4.clone  # check that python tdi save is the same as labview tdi save
    t2.save(path.join(local, "mixedmetatest2.txt"))
    t2l = Data(path.join(local, "mixedmetatest2.txt"))
    for orig, load in [(t, tl), (t2, t2l)]:
        for k in ["Loaded as", "TDI Format"]:
            orig[k] = load[k]
        assert np.allclose(orig.data, load.data)
        assert orig.column_headers == load.column_headers
        _ = load.metadata ^ orig.metadata
        assert load.metadata == orig.metadata, "Metadata not the same on round tripping to disc"
    # os.remove(path.join(local, "mixedmetatest.txt")) #clear up
    # os.remove(path.join(local, "mixedmetatest2.txt"))


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec  8 15:20:15 2018

@author: phygbu
"""

import pytest
import os.path as path
from Stoner.core.base import typeHintedDict

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
# sys.path.insert(0,pth)


def test_ops():
    d = typeHintedDict([("el1", 1), ("el2", 2), ("el3", 3), ("other", 4)])
    d.filter("el")
    assert len(d) == 3
    d.filter(lambda x: x.endswith("3"))
    assert len(d) == 1
    assert d["el3"] == 3
    d["munge"] = None
    assert d.types["munge"] == "Void", "Setting type for None value failed."
    d["munge"] = 1
    assert d["munge{String}"] == "1", "Munging return type in getitem failed."
    assert (
        repr(d)
        == """'el3':I32:3
'munge':I32:1"""
    ), "Repr failed \n{}".format(d)
    assert len(d | {"test": 2}) == len(d) + 1
    assert len(d | {"munge": 2}) == len(d)
    e = d.copy()
    e |= {"test": 4}
    assert d != e
    d.update({"test": 4})
    assert d == e


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""":py:class:`lmfit.Model` model classes and functions for various superconductivity related models."""
# pylint: disable=invalid-name
# This module can be used with Stoner v.0.9.0 asa standalone module
__all__ = [
    "RSJ_Noiseless",
    "RSJ_Simple",
    "Strijkers",
    "Ic_B_Airy",
    "rsj_noiseless",
    "rsj_simple",
    "strijkers",
    "ic_B_airy",
]

from functools import partial

import numpy as np
from scipy.special import jv
from scipy.constants import physical_constants
from scipy.integrate import quad

from lmfit import Model
from lmfit.models import update_param_vals

hbar = physical_constants["Planck constant over 2 pi"]
kb = physical_constants["Boltzmann constant"]
Phi_0 = physical_constants["mag. flux quantum"][0]

J1 = partial(jv, 1)


try:  # numba is an optional dependency
    from numba import jit, float64
except ImportError:
    from ....compat import _dummy, _jit as jit

    float64 = _dummy()


@jit(float64[:](float64[:], float64, float64, float64, float64), nopython=True, parallel=True, nogil=True)
def _strijkers_core(V, omega, delta, P, Z):
    """Implement strijkers Model for point-contact Andreev Reflection Spectroscopy.

    Args:
        V = bias voltages, params=list of parameter values, imega, delta,P and Z
        omega (float): Broadening
        delta (float): SC energy Gap
        P (float): Interface parameter
        Z (float): Current spin polarization through contact

    Return:
        Conductance vs bias data.

    Note:
           PCAR fitting Strijkers modified BTK model TK PRB 25 4515 1982, Strijkers PRB 63, 104510 2000

    This version only uses 1 delta, not modified for proximity
    """
    #   Parameters

    mv = np.max(np.abs(V))  # Limit for evaluating the integrals
    E = np.linspace(-2 * mv, 2 * mv, V.size * 20)  # Energy range in meV - we use a mesh 20x denser than data points
    gauss = np.exp(-(E**2 / (2 * omega**2)))
    gauss /= gauss.sum()  # Normalised gaussian for the convolution

    # Conductance calculation
    #    For ease of calculation, epsilon = E/(sqrt(E^2 - delta^2))
    #    Calculates reflection probabilities when E < or > delta
    #    A denotes Andreev Reflection probability
    #    B denotes normal reflection probability
    #    subscript p for polarised, u for unpolarised
    #    Ap is always zero as the polarised current has 0 prob for an Andreev
    #    event

    Au1 = (delta**2) / ((E**2) + (((delta**2) - (E**2)) * (1 + 2 * (Z**2)) ** 2))
    Au2 = (((np.abs(E) / (np.sqrt((E**2) - (delta**2)))) ** 2) - 1) / (
        ((np.abs(E) / (np.sqrt((E**2) - (delta**2)))) + (1 + 2 * (Z**2))) ** 2
    )
    Bu2 = (4 * (Z**2) * (1 + (Z**2))) / (((np.abs(E) / (np.sqrt((E**2) - (delta**2)))) + (1 + 2 * (Z**2))) ** 2)
    Bp2 = Bu2 / (1 - Au2)

    unpolarised_prefactor = (1 - P) * (1 + (Z**2))
    polarised_prefactor = 1 * (P) * (1 + (Z**2))
    # Optimised for a single use of np.where
    G = (
        unpolarised_prefactor
        + polarised_prefactor
        + +np.where(
            np.abs(E) <= delta,
            unpolarised_prefactor * (2 * Au1 - 1) - np.ones_like(E) * polarised_prefactor,
            unpolarised_prefactor * (Au2 - Bu2) - Bp2 * polarised_prefactor,
        )
    )

    # Convolve and chop out the central section
    cond = np.convolve(G, gauss)
    cond = cond[(E.size // 2) : 3 * (E.size // 2)]
    # Linear interpolation back onto the V data point
    matches = np.searchsorted(E, V)
    condl = cond[matches - 1]
    condh = cond[matches]
    El = E[matches - 1]
    Er = E[matches]
    cond = (condh - condl) / (Er - El) * (V - El) + condl
    return cond


def strijkers(V, omega, delta, P, Z):
    """Strijkers Model for point-contact Andreev Reflection Spectroscopy.

    Args:
        V (array):
            bias voltages
        omega (float):
            Broadening
        delta (float):
            SC energy Gap
        P (float):
            Interface parameter
        Z (float):
            Current spin polarization through contact

    Return:
        Conductance vs bias data.

    .. note::

       PCAR fitting Strijkers modified BTK model TK PRB 25 4515 1982, Strijkers PRB 63, 104510 2000

    This version only uses 1 delta, not modified for proximity

    Example:
        .. plot:: samples/lmfit_demo.py
            :include-source:
            :outname: strijkers_func
    """
    if isinstance(V, np.ma.MaskedArray):
        mask = V.mask
        V = np.array(V)
    else:
        mask = False
    ret = _strijkers_core(V, omega, delta, P, Z)
    ret = np.ma.MaskedArray(ret)
    ret.mask = mask
    return ret


def rsj_noiseless(I, Ic_p, Ic_n, Rn, V_offset):
    r"""Implement a simple noiseless RSJ model.

    Args:
        I (array-like): Current values
        Ic_p (foat): Critical current on positive branch
        Ic_n (foat): Critical current on negative branch
        Rn (float): Normal state resistance
        V_offset(float): Offset volage in measurement

    Returns:
        (array) Calculated voltages

    Notes:
        Impleemtns a simple form of the RSJ model for a Josephson Junction:

            :math:`V(I)=R_N\frac{I}{|I|}\sqrt{I^2-I_c^2}-V_{offset}`

    Example:
        .. plot:: samples/Fitting/rsj_fit.py
            :include-source:
            :outname: rsj_noiseless_func
    """
    normal_p = np.sign(I) * np.real(np.sqrt(I**2 - Ic_p**2)) * Rn
    normal_n = np.sign(I) * np.real(np.sqrt(I**2 - Ic_n**2)) * Rn
    p_branch = np.where(I > Ic_p, normal_p, np.zeros_like(I))
    n_branch = np.where(I < Ic_n, normal_n, p_branch)
    return n_branch + V_offset


def rsj_simple(I, Ic, Rn, V_offset):
    r"""Implement a simple noiseless symmetric RSJ model.

    Args:
        I (array-like):
            Current values
        Ic (foat):
            Critical current
        Rn (float):
            Normal state resistance
        V_offset(float):
            Offset volage in measurement

    Returns:
        (array):
            Calculated voltages

    Notes:
        Impleemtns a simple form of the RSJ model for a Josephson Junction:

            :math:`V(I)=R_N\frac{I}{|I|}\sqrt{I^2-I_c^2}-V_{offset}`

    Example:
        .. plot:: samples/Fitting/rsj_fit.py
            :include-source:
            :outname: rsj_simple_func
    """
    normal = Rn * np.sign(I) * np.real(np.sqrt(I**2 - Ic**2))
    ic_branch = np.zeros_like(I)
    return np.where(np.abs(I) < Ic, ic_branch, normal) + V_offset


def ic_B_airy(B, Ic0, B_offset, A):
    r"""Calculate Critical Current for a round Josepshon Junction wrt to Field.

    Args:
        B (array-like):
            Magnetic Field (structly flux density in T)
        Ic0 (float):
            Maximum critical current
        B_offset (float):
            Field offset/trapped flux in coils/remanent M in junction
        A(fl,oat):
            Area of junction in $m^2$

    Returns:
        (array):
            Values of critical current

    Notes:
        Represents the critical current as:
            :math:`I_{c0}\times\left|\frac{2 J_1\left(\frac{\pi\(B-B_{offset}) A}\right)}{\Phi_0}}
                    {\frac{\pi\(B-B_{offset}) A}){\Phi_0}}\right|`
        where :math:`J_1` is a first order Bessel function.

        For small ($<1^{-5}$)values of the Bessel function argument, this will return Ic0 to
        ensure correct evaluation for 0 flux.

    Example:
        .. plot:: samples/Fitting/ic_b_airy.py
            :include-source:
            :outname: ic_b_airy_func
    """
    arg = (B - B_offset) * A * np.pi / Phi_0

    return Ic0 * np.abs(2 * np.where(np.abs(arg) < 1e-5, np.ones_like(arg), J1(arg) / arg))


def icRN_Clean(d_f, IcRn0, E_x, v_f, d_0):
    r"""Critical Current versus ferromagnetic narrier thickness, clean limit.

    Args:
        d_f (array):
            ferromagnetic barrier thickness (nm)
        IcRn0 (float):
            Characteristic voltage scaling factor
        E_x (float):
            Exchange energy (eV)
        v_f (float):
            Fermi velocity (ms^-1)
        d_0 (float):
            barrier thickness offset

    Returns:
        (array):
            IcRn values

    Notes:
        Implements

        Lmath:`I_cR_N = I_cR_N^0\zfrac{\sin(2E_x (d_f-d_0)/hv_f)}{2E_x(d_f-d_0)/hv_f}`
    """
    h = physical_constants["Planck constant in eV s"]
    x = (d_f - d_0) * 1e-9
    A = 2 * E_x / (v_f * h)
    return IcRn0 * np.abs(np.sin(A * x)) / (A * x)


def ic_RN_Dirty(d_f, IcRn0, E_x, v_f, d_0, tau, delta, T):
    r"""Critical Current versus ferromagnetic narrier thickness, clean limit.

    Args:
        d_f (array):
            ferromagnetic barrier thickness (nm)
        IcRn0 (float):
            Characteristic voltage scaling factor
        E_x (float):
            Exchange energy (eV)
        v_f (float):
            Fermi velocity (ms^-1)
        d_0 (float):
            barrier thickness offset
        l (float):
            mean-free-path (nm)

    Returns:
        (array):
            IcRn values

    Notes:
        Implements Eq 18 from F.S. Bergeret, A.F. Volkov, and K.B. Efetov, Phys. Rev. B 64, 134506 (2001).
    """
    L = v_f * tau * 1e-9
    t = tau / hbar
    w_m = lambda m: (2 * m + 1) * T * np.pi * kb
    k_m = lambda m: (1 + 2 * np.abs(w_m(m)) * t) + (0 - 2j) * E_x * t
    integrad = lambda mu, m: np.real(mu / (np.sinh(d_f * k_m(m) / (mu * L))))
    prefactor = lambda m: delta**2 / (delta**2 + w_m(m) ** 2)

    term = lambda m: prefactor(m) * quad(integrad, -1, 1, (m,))  # pylint: disable=W0612


class Strijkers(Model):
    """strijkers Model for point-contact Andreev Reflection Spectroscopy.

    Args:
        V (array): bias voltages
        omega (float): Broadening
        delta (float): SC energy Gap
        P (float): Interface parameter
        Z (float): Current spin polarization through contact

    Return:
        Conductance vs bias data.

    .. note::

       PCAR fitting Strijkers modified BTK model TK PRB 25 4515 1982, Strijkers PRB 63, 104510 2000

    This version only uses 1 delta, not modified for proximity

    Example:
        .. plot:: samples/lmfit_demo.py
            :include-source:
            :outname: strijkers_class
    """

    display_names = [r"\omega", r"\Delta", "P", "Z"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(strijkers, *args, **kwargs)

    def guess(self, data, **kwargs):  # pylint: disable=unused-argument
        """Guess starting values for a good Nb contact to a ferromagnet at 4.2K."""
        pars = self.make_params(omega=0.5, delta=1.50, P=0.42, Z=0.15)
        pars["omega"].min = 0.36
        pars["omega"].max = 5.0
        pars["delta"].min = 0.5
        pars["delta"].max = 2.0
        pars["Z"].min = 0.1
        pars["P"].min = 0.0
        pars["P"].max = 1.0
        return update_param_vals(pars, self.prefix, **kwargs)


class RSJ_Noiseless(Model):
    r"""Implement a simple noiseless RSJ model.

    Args:
        I (array-like): Current values
        Ic_p (foat): Critical current on positive branch
        Ic_n (foat): Critical current on negative branch
        Rn (float): Normal state resistance
        V_offset(float): Offset volage in measurement

    Returns:
        (array) Calculated voltages

    Notes:
        Impleemtns a simple form of the RSJ model for a Josephson Junction:

            :math:`V(I)=R_N\frac{I}{|I|}\sqrt{I^2-I_c^2}-V_{offset}`

    Example:
        .. plot:: samples/Fitting/rsj_fit.py
            :include-source:
            :outname: rsj_noiseless_class
    """

    display_names = ["I_c^p", "I_c^n", "R_N", "V_{offset}"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(rsj_noiseless, *args, **kwargs)

    def guess(self, data, **kwargs):
        """Guess parameters as gamma=2, H_k=0, M_s~(pi.f)^2/(mu_0^2.H)-H."""
        x = kwargs.get("x", np.linspace(1, len(data), len(data) + 1))

        v_offset_guess = np.mean(data)
        v = np.abs(data - v_offset_guess)
        x = np.abs(x)

        v_low = np.max(v) * 0.05
        v_high = np.max(v) * 0.90

        ic_index = v < v_low
        rn_index = v > v_high
        ic_guess = np.max(x[ic_index])  # Guess Ic from a 2% of max V threhsold creiteria

        rn_guess = np.mean(v[rn_index] / x[rn_index])

        pars = self.make_params(Ic_p=ic_guess, Ic_n=-ic_guess, Rn=rn_guess, V_offset=v_offset_guess)
        pars["Ic_p"].min = 0
        pars["Ic_n"].max = 0
        return update_param_vals(pars, self.prefix, **kwargs)


class RSJ_Simple(Model):
    r"""Implements a simple noiseless symmetric RSJ model.

    Args:
        I (array-like): Current values
        Ic (foat): Critical current
        Rn (float): Normal state resistance
        V_offset(float): Offset volage in measurement

    Returns:
        (array) Calculated voltages

    Notes:
        Impleemtns a simple form of the RSJ model for a Josephson Junction:

            :math:`V(I)=R_N\frac{I}{|I|}\sqrt{I^2-I_c^2}-V_{offset}`

    Example:
        .. plot:: samples/Fitting/rsj_fit.py
            :include-source:
            :outname: rsj_simple_class

    """

    display_names = ["I_c", "R_N", "V_{offset}"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(rsj_simple, *args, **kwargs)

    def guess(self, data, **kwargs):
        """Guess parameters as gamma=2, H_k=0, M_s~(pi.f)^2/(mu_0^2.H)-H."""
        x = kwargs.get("x", np.linspace(1, len(data), len(data) + 1))

        v_offset_guess = np.mean(data)
        v = np.abs(data - v_offset_guess)
        x = np.abs(x)

        v_low = np.max(v) * 0.05
        v_high = np.max(v) * 0.90

        ic_index = v < v_low
        rn_index = v > v_high
        ic_guess = np.max(x[ic_index])  # Guess Ic from a 2% of max V threhsold creiteria

        rn_guess = np.mean(v[rn_index] / x[rn_index])

        pars = self.make_params(Ic=ic_guess, Rn=rn_guess, V_offset=v_offset_guess)
        # pars["Ic"].min = 0
        return update_param_vals(pars, self.prefix, **kwargs)


class Ic_B_Airy(Model):
    r"""Critical Current for a round Josepshon Junction wrt to Field.

    Args:
        B (array-like):
            Magnetic Field (structly flux density in T)
        Ic0 (float):
            Maximum critical current
        B_offset (float):
            Field offset/trapped flux in coils/remanent M in junction
        A(fl,oat):
            Area of junction in $m^2$

    Returns:
        (array):
            Values of critical current

    Notes:
        Represents the critical current as:
            :math:`I_{c0}\times\left|\frac{2 J_1\left(\frac{\pi\(B-B_{offset}) A}\right)}
                        {\Phi_0}}{\frac{\pi\(B-B_{offset}) A}){\Phi_0}}\right|`
        where `J_1` is a first order Bessel function.

    Example:
        .. plot:: samples/Fitting/ic_b_airy.py
            :include-source:
            :outname: ic_b_airy_class

    """

    display_names = ["I_{c0}", "B_{offset}"]

    def __init__(self, *args, **kwargs):
        """Configure Initial fitting function."""
        super().__init__(ic_B_airy, *args, **kwargs)

    def guess(self, data, **kwargs):
        """Guess parameters as max(data), x[argmax(data)] and from FWHM of peak."""
        x = kwargs.get("x", np.linspace(-len(data) / 2, len(data) / 2, len(data)))

        Ic0_guess = data.max()
        B_offset_guess = x[data.argmax()]
        tmp = np.abs(data - (data.max() / 2))
        x0 = np.abs(x[tmp.argmin()] - B_offset_guess)
        A_guess = 2.2 * Phi_0 / (np.pi * x0)

        pars = self.make_params(Ic0=Ic0_guess, B_offset=B_offset_guess, A=A_guess)
        pars["Ic0"].min = 0
        return update_param_vals(pars, self.prefix, **kwargs)



"""Example of nDimArrhenius Fit."""

# pylint: disable=invalid-name
from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.tunnelling import simmons, Simmons

# Make some data
V = linspace(-4, 4, 101)
I = simmons(V, 2500, 3.2, 15.0) + normal(size=len(V), scale=5e-7)
dI = ones_like(V) * 500e-9

p0 = p0 = [2500, 3, 10.0]

d = Data(V, I, dI, setas="xye", column_headers=["Bias", "Current", "Noise"])

d.curve_fit(simmons, p0=p0, result=True, header="curve_fit", maxfev=2000)
d.setas = "xyey"
d.plot(fmt=["r,", "b-"], capsize=1)
d.annotate_fit(
    simmons,
    x=0.25,
    y=0.25,
    prefix="simmons",
    fontdict={"size": "x-small", "color": "blue"},
)

d.setas = "xye"
fit = Simmons()
d.lmfit(Simmons, p0=p0, result=True, header="lmfit", maxfev=2000)
d.setas = "x...y"
d.plot(fmt="g-", label="lmfit")
d.annotate_fit(
    fit,
    x=0.65,
    y=0.25,
    prefix="Simmons",
    fontdict={"size": "x-small", "color": "green"},
)

d.ylabel = "Current (A)"
d.xlabel = "Bias (V)"
d.title = "Simmons Model test"
d.yscale("symlog", linthresh=1e-5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Demo of the make_model decorator."""
# pylint: disable=invalid-name, redefined-outer-name
from numpy import linspace
from numpy.random import normal, seed

from Stoner import Data
from Stoner.analysis.fitting.models import make_model

seed(12345)  # Ensure consistent random numbers


# Make our model
@make_model
def simple_model(x, m, c):
    """A straight line."""
    return x * m + c


# Add a function to guess parameters (optional, but)
@simple_model.guesser
def guess_vals(y, x=None):
    """Should guess parameter values really!"""
    m = (y.max() - y.min()) / (x[y.argmax()] - x[y.argmin()])
    c = x.mean() * m - y.mean()  # return one value per parameter
    return [m, c]


# Add a function to sry vonstraints on parameters (optional)
@simple_model.hinter
def hint_parameters():
    """Five some hints about the parameter."""
    return {"m": {"max": 10.0, "min": 0.0}, "c": {"max": 5.0, "min": -5.0}}


# Create some x,y data
x = linspace(0, 10, 101)
y = 4.5 * x - 2.3 + normal(scale=0.4, size=len(x))

# Make The Data object
d = Data(x, y, setas="xy", column_headers=["X", "Y"])

# Do the fit
d.lmfit(simple_model, result=True)

# Plot the result
d.setas = "xyy"
d.plot(fmt=["r+", "b-"])
d.title = "Simple Model Fit"
d.annotate_fit(simple_model, x=0.05, y=0.5)



import os.path as path
import os
import runpy
from Stoner.compat import listdir_recursive
import matplotlib.pyplot as plt
from traceback import format_exc
import pytest
import warnings

warnings.filterwarnings("ignore")

pth = path.dirname(__file__)
pth = path.realpath(path.join(pth, "../../"))
datadir = path.join(pth, "doc", "samples")


def get_scripts():
    skip_scipts = ["plot_folder_demo"]
    scripts = [path.realpath(x) for x in listdir_recursive(datadir, "*.py") if not x.endswith("__init__.py")]
    scripts = {
        path.splitext(path.basename(x))[0]: x
        for x in scripts
        if path.splitext(path.basename(x))[0].lower() not in skip_scipts
    }
    scripts = list(dict(sorted(scripts.items())).values())
    return scripts


@pytest.mark.parametrize("script", get_scripts())
@pytest.mark.filterwarnings("ignore:.*:RuntimeWarning")
def test_scripts(script, monkeypatch):
    """Import each of the sample scripts in turn and see if they ran without error"""
    print(f"Trying script {script}")
    try:
        os.chdir(datadir)
        runpy.run_path(script)
        fignum = len(plt.get_fignums())
        assert fignum >= 1, f"{script} Did not produce any figures !"
        print("Done")
        plt.close("all")
    except Exception:
        error = format_exc()
        print(f"Failed with\n{error}")
        assert False, f"Script {script} failed with {error}"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main(["--pdb", __file__])



# -*- coding: utf-8 -*-
"""Creat File dialog boxes using the PyQt5 module.

Code based on the PyQt5 Tutorial code,
"""
__all__ = ["fileDialog"]
import pathlib
from typing import Any, Union, Optional, Dict, Type
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import SpanSelector
from matplotlib.patches import Rectangle

from ..compat import mpl_version

try:
    from PyQt5.QtWidgets import QWidget, QFileDialog, QApplication
except ImportError:

    class App:
        """Mock App that raises an error when you try to call openDialog on it."""

        modes: Dict = dict()

        def openDialog(
            self,
            title: Optional[str] = None,
            start: Union[str, pathlib.Path] = "",
            patterns: Optional[Dict] = None,
            mode: str = "OpenFile",
        ) -> Optional[pathlib.Path]:
            """Raise and error because PyQT5 not present."""
            raise ValueError("Cammpt open a dialog box because PqQt5 library missing!")

else:

    class App(QApplication):
        """Placehold PyQT5 Application for producing filedialog boxes."""

        modes = {
            "OpenFile": {
                "method": QFileDialog.getOpenFileName,
                "caption": "Select file to open...",
                "arg": ["parent", "caption", "directory", "filter", "options"],
            },
            "OpenFiles": {
                "method": QFileDialog.getOpenFileNames,
                "caption": "Select file(s_ to open...",
                "arg": ["parent", "caption", "directory", "filter", "options"],
            },
            "SaveFile": {
                "method": QFileDialog.getSaveFileName,
                "caption": "Save file as...",
                "arg": ["parent", "caption", "directory", "filter", "options"],
            },
            "SelectDirectory": {
                "method": QFileDialog.getExistingDirectory,
                "caption": "Select folder...",
                "arg": ["parent", "caption", "directory", "options"],
            },
        }

        def __init__(self, *args: Any, **kargs: Any) -> None:
            super().__init__([""], *args, **kargs)
            self.title = "PyQt5 file dialogs - pythonspot.com"
            self.left = 10
            self.top = 10
            self.width = 640
            self.height = 480
            self.initUI()

        def initUI(self) -> None:
            self.dialog = QWidget()
            self.dialog.title = "PyQt5 file dialogs - pythonspot.com"
            self.dialog.left = 10
            self.dialog.top = 10
            self.dialog.width = 640
            self.dialog.height = 480

            self.dialog.setWindowTitle(self.title)
            self.dialog.setGeometry(self.left, self.top, self.width, self.height)

        def openDialog(
            self,
            title: Optional[str] = None,
            start: Union[str, pathlib.Path] = "",
            patterns: Optional[Dict] = None,
            mode: str = "OpenFile",
        ) -> Optional[pathlib.Path]:
            """Create a dialog box for selecting filenames or directories.

            Keyword Arguments:
                title (str, None):
                    Label of the dialog box, default None will select something depending on the mode.
                start (str):
                    The starting directory for the dialog box.
                patterns (dict):
                    Filename patterns - the keys of the dictionary are glob patterns, the values the corresponding
                    explanation of the file
                    type.
                mode (str):
                    Determines the type of filedialog box used. Values are:
                        -   "OpenFile" - get a single existing file
                        -   "OpenFiles" - get one or more existing files
                        -   "SaveFile" - get an existing or new file - warns about overwriting files
                        -   "SelectDirectory" - gets the name of an existing (possibly newly created) directory.

            Returns:
                (str, None):
                    Either a string containing the absolute path to the file or directory, or None if the dialog
                    was cancelled.
            """
            if mode not in self.modes:
                raise ValueError(f"Unknown dialog mode {mode}")
            method = self.modes[mode]["method"]
            if title is None:
                title = self.modes[mode]["caption"]
            if patterns is None:
                patterns = {"*.*": "All Files", "*.py": "Python Files"}
            patterns = ";;".join([f"{v} ({k})" for k, v in patterns.items()])
            options = QFileDialog.Options()

            kwargs = {"caption": title, "directory": str(start), "filter": patterns, "options": options, "modal": True}
            kwargs = {k: kwargs[k] for k in (set(kwargs.keys()) & set(self.modes[mode]["arg"]))}

            ret = method(self.dialog, **kwargs)

            if isinstance(ret, tuple):
                ret = ret[0]
            if isinstance(ret, (str, pathlib.PurePath)):
                ret = pathlib.Path(ret)
            elif isinstance(ret, list):
                ret = [pathlib.Path(x) for x in ret]
            elif not ret:
                ret = None
            else:
                raise TypeError(f"Something when wrong here - can't handle {ret} as a {type(ret)}")
            return ret


class RangeSelect:
    """A simple class to allow a matplotlib graph to be used to select data."""

    def __init__(self):
        """Initialise the selector state."""
        self.data = None
        self.finished = False
        self.selector = []
        self.invert = False
        self.xcol = None
        self.ycol = None
        self.selection = []
        self.ax = None

    def __call__(self, data, xcol, accuracy, invert=False):
        """Run the selector with the data."""
        self.data = data
        self.data._select = self  # To allow for unit testing
        self.xcol = xcol
        self.invert = invert

        col = "red" if self.invert else "green"

        if len(self.data.setas.y) > 0:
            self.ycol = self.data.setas.y
        else:
            self.ycol = list(range(self.data.shape[1]))
            self.ycol.remove(self.xcol)
        # Preserve figure settings before creating plot
        fig_tmp = getattr(self.data, "fig", None), getattr(self.data, "axes", None)
        fig = plt.figure()
        self.data.plot(self.xcol, self.ycol, figure=fig)
        self.data.title = "Select Data and press Enter to confirm\nEsc to cancel, i to invert selection."
        self.ax = self.data.axes[-1]
        if mpl_version.minor >= 5:
            kargs = {"props": {"edgecolor": col, "facecolor": col, "alpha": 0.5}}
        else:
            kargs = {"rectprops": {"edgecolor": col, "facecolor": col, "alpha": 0.5}}
        self.selector = SpanSelector(
            self.ax,
            self.onselect,
            "horizontal",
            useblit=True,
            **kargs,
        )
        fig.canvas.mpl_connect("key_press_event", self.keypress)
        while not self.finished:
            plt.pause(0.1)
        # Clean up and restore the figure settings
        plt.close(self.data.fig.number)
        if fig_tmp[0] is not None and fig_tmp[0] in plt.get_fignums():
            self.data.fig = fig_tmp[0]
        delattr(self.data, "_select")
        idx = np.ones(len(self.data), dtype=bool)
        for ix, selection in enumerate(self.selection):
            if ix == 0:
                idx = self.data._search_index(xcol, selection, accuracy, invert=self.invert)
            else:
                idx = np.logical_or(idx, self.data._search_index(xcol, selection, accuracy, invert=self.invert))
        return idx

    def onselect(self, xmin, xmax):
        """Add the selection limits to the selections list."""
        self.selection.append((xmin, xmax))
        ylim = self.ax.get_ylim()
        col = "red" if self.invert else "green"
        rect = Rectangle((xmin, ylim[0]), xmax - xmin, ylim[1] - ylim[0], edgecolor=col, facecolor=col, alpha=0.5)
        self.ax.add_patch(rect)

    def keypress(self, event):
        """Habndle key press events.

        Args:
            event (matplotlib event):
                The matplotlib event object

        Returns:
            None.

        Tracks the mode and adjusts the cursor display.
        """
        if event.key.lower() == "enter":  # Finish selection
            self.finished = True
        elif event.key.lower() == "escape":  # Abandon selection
            self.selection = []
            self.finished = True
        elif event.key.lower() == "backspace":  # Delete last selection
            if len(self.selection) > 0:
                del self.selection[-1]
                self.ax.patches[-1].remove()
                self.data.fig.canvas.draw()
        elif event.key.lower() == "i":  # Invert selection
            self.invert = ~self.invert
            col = "red" if self.invert else "green"
            for p in self.ax.patches:
                p.update({"color": col, "facecolor": col})


fileDialog: Type[App] = App()



"""Stoner.Folders : Classes for working collections of :class:`.Data` files.

The core classes provides a means to access them as an ordered collection or as a mapping.
"""

__all__ = ["DataFolder", "PlotFolder"]

from Stoner.tools import make_Data
from .folders.core import baseFolder

from .folders.mixins import DiskBasedFolderMixin, DataMethodsMixin, PlotMethodsMixin


class DataFolder(DataMethodsMixin, DiskBasedFolderMixin, baseFolder):
    """Provide an interface to manipulating lots of data files stored within a directory structure on disc.

    By default, the members of the DataFolder are instances of :class:`Stoner.Data`. The DataFolder emplys a lazy
    open strategy, so that files are only read in from disc when actually needed.

    .. inheritance-diagram:: DataFolder

    """

    def __init__(self, *args, **kargs):
        """Set the default type before creating the DataFolder."""
        self.type = kargs.pop("type", make_Data(None))
        super().__init__(*args, **kargs)


class PlotFolder(PlotMethodsMixin, DataFolder):
    """A :py:class:`Stoner.folders.baseFolder` that knows how to ploth its underlying data files."""
